# Copyright 2014-2017 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Functions for graphical output."""

from __future__ import print_function, division, absolute_import
import numpy as np
import warnings

from odl.core.util.testutils import run_doctests
from odl.core.util.dtype_utils import is_real_dtype
from odl.core.array_API_support import get_array_and_backend

__all__ = ('show_discrete_data',)


def warning_free_pause():
    """Issue a matplotlib pause without the warning."""
    import matplotlib.pyplot as plt

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore",
                                message="Using default event loop until "
                                        "function specific to this GUI is "
                                        "implemented")
        plt.pause(0.0001)


def _safe_minmax(values):
    """Calculate min and max of array with guards for nan and inf."""

    # Nan and inf guarded min and max
    isfinite = np.isfinite(values)
    if np.any(isfinite):
        # Only use finite values
        values = values[isfinite]

    minval = np.min(values)
    maxval = np.max(values)

    return minval, maxval


def _colorbar_ticks(minval, maxval):
    """Return the ticks (values show) in the colorbar."""
    if not (np.isfinite(minval) and np.isfinite(maxval)):
        return [0, 0, 0]
    elif minval == maxval:
        return [minval]
    else:
        # Add eps to ensure values stay inside the range of the colorbar.
        # Otherwise they may occationally not display.
        eps = (maxval - minval) / 1e5
        return [minval + eps, (maxval + minval) / 2., maxval - eps]


def _digits(minval, maxval):
    """Digits needed to comforatbly display values in [minval, maxval]"""
    if minval == maxval:
        return 3
    else:
        return min(10, max(2, int(1 + abs(np.log10(maxval - minval)))))


def _colorbar_format(minval, maxval):
    """Return the format string for the colorbar."""
    if not (np.isfinite(minval) and np.isfinite(maxval)):
        return str(maxval)
    else:
        return '%.{}f'.format(_digits(minval, maxval))


def _axes_info(grid, npoints=5):
    result = []

    min_pt = grid.min()
    max_pt = grid.max()
    for axis in range(grid.ndim):
        xmin = min_pt[axis]
        xmax = max_pt[axis]

        points = np.linspace(xmin, xmax, npoints)
        indices = np.linspace(0, grid.shape[axis] - 1, npoints, dtype=int)
        tick_values = grid.coord_vectors[axis][indices]

        # Do not use corner point in case of a partition, use outer corner
        tick_values[[0, -1]] = xmin, xmax

        format_str = '{:.' + str(_digits(xmin, xmax)) + 'f}'
        tick_labels = [format_str.format(f) for f in tick_values]

        result += [(points, tick_labels)]

    return result


def show_discrete_data(values, grid, title=None, method='',
                       force_show=False, fig=None, **kwargs):
    """Display a discrete 1d or 2d function.

    Parameters
    ----------
    values : `numpy.ndarray`
        The values to visualize.

    grid : `RectGrid` or `RectPartition`
        Grid of the values.

    title : string, optional
        Set the title of the figure.

    method : string, optional
        1d methods:

        'plot' : graph plot

        'scatter' : scattered 2d points
        (2nd axis <-> value)

        2d methods:

        'imshow' : image plot with coloring according to value,
        including a colorbar.

        'scatter' : cloud of scattered 3d points
        (3rd axis <-> value)

        'wireframe', 'plot_wireframe' : surface plot

    force_show : bool, optional
        Whether the plot should be forced to be shown now or deferred until
        later. Note that some backends always displays the plot, regardless
        of this value.

    fig : `matplotlib.figure.Figure`, optional
        The figure to show in. Expected to be of same "style", as the figure
        given by this function. The most common usecase is that fig is the
        return value from an earlier call to this function.
        Default: New figure

    interp : {'nearest', 'linear'}, optional
        Interpolation method to use.
        Default: 'nearest'

    axis_labels : string, optional
        Axis labels, default: ['x', 'y']

    update_in_place : bool, optional
        Update the content of the figure in-place. Intended for faster real
        time plotting, typically ~5 times faster.
        This is only performed for ``method == 'imshow'`` with real data and
        ``fig != None``. Otherwise this parameter is treated as False.
        Default: False

    axis_fontsize : int, optional
        Fontsize for the axes. Default: 16

    colorbar : bool, optional
        Argument relevant for 2d plots using ``method='imshow'``. If ``True``,
        include a colorbar in the plot.
        Default: True

    kwargs : {'figsize', 'saveto', ...}, optional
        Extra keyword arguments passed on to display method
        See the Matplotlib functions for documentation of extra
        options.

    Returns
    -------
    fig : `matplotlib.figure.Figure`
        The resulting figure. It is also shown to the user.

    See Also
    --------
    matplotlib.pyplot.plot : Show graph plot

    matplotlib.pyplot.imshow : Show data as image

    matplotlib.pyplot.scatter : Show scattered 3d points
    """
    # Importing pyplot takes ~2 sec, only import when needed.
    import matplotlib.pyplot as plt

    args_re = []
    args_im = []
    dsp_kwargs = {}
    sub_kwargs = {}
    arrange_subplots = (121, 122)  # horzontal arrangement

    # Create axis labels which remember their original meaning
    axis_labels = kwargs.pop('axis_labels', ['x', 'y'])

    values_are_complex = not is_real_dtype(values.dtype)
    figsize = kwargs.pop('figsize', None)
    saveto = kwargs.pop('saveto', None)
    interp = kwargs.pop('interp', 'nearest')
    axis_fontsize = kwargs.pop('axis_fontsize', 16)
    colorbar = kwargs.pop('colorbar', True)

    # Normalize input
    interp, interp_in = str(interp).lower(), interp
    method, method_in = str(method).lower(), method

    # Check if we should and can update the plot in-place
    update_in_place = kwargs.pop('update_in_place', False)
    if (update_in_place and
            (fig is None or values_are_complex or values.ndim != 2 or
             (values.ndim == 2 and method not in ('', 'imshow')))):
        update_in_place = False

    if values.ndim == 1:  # TODO: maybe a plotter class would be better
        if not method:
            if interp == 'nearest':
                method = 'step'
                dsp_kwargs['where'] = 'mid'
            elif interp == 'linear':
                method = 'plot'
            else:
                raise ValueError('`interp` {!r} not supported'
                                 ''.format(interp_in))

        if method == 'plot' or method == 'step' or method == 'scatter':
            args_re += [grid.coord_vectors[0], values.real]
            args_im += [grid.coord_vectors[0], values.imag]
        else:
            raise ValueError('`method` {!r} not supported'
                             ''.format(method_in))

    elif values.ndim == 2:
        if not method:
            method = 'imshow'

        if method == 'imshow':
            values, array_backend = get_array_and_backend(values)
            values = array_backend.to_numpy(values)
            args_re = [np.rot90(values.real)]
            args_im = [np.rot90(values.imag)] if values_are_complex else []

            extent = [grid.min()[0], grid.max()[0],
                      grid.min()[1], grid.max()[1]]

            if interp == 'nearest':
                interpolation = 'nearest'
            elif interp == 'linear':
                interpolation = 'bilinear'
            else:
                raise ValueError('`interp` {!r} not supported'
                                 ''.format(interp_in))

            dsp_kwargs.update({'interpolation': interpolation,
                               'cmap': 'bone',
                               'extent': extent,
                               'aspect': 'auto'})
        elif method == 'scatter':
            pts = grid.points()
            args_re = [pts[:, 0], pts[:, 1], values.ravel().real]
            args_im = ([pts[:, 0], pts[:, 1], values.ravel().imag]
                       if values_are_complex else [])
            sub_kwargs.update({'projection': '3d'})
        elif method in ('wireframe', 'plot_wireframe'):
            method = 'plot_wireframe'
            x, y = grid.meshgrid
            args_re = [x, y, np.rot90(values.real)]
            args_im = ([x, y, np.rot90(values.imag)] if values_are_complex
                       else [])
            sub_kwargs.update({'projection': '3d'})
        else:
            raise ValueError('`method` {!r} not supported'
                             ''.format(method_in))

    else:
        raise NotImplementedError('no method for {}d display implemented'
                                  ''.format(values.ndim))

    # Additional keyword args are passed on to the display method
    dsp_kwargs.update(**kwargs)

    if fig is not None:
        # Reuse figure if given as input
        if not isinstance(fig, plt.Figure):
            raise TypeError('`fig` {} not a matplotlib figure'.format(fig))

        if not plt.fignum_exists(fig.number):
            # If figure does not exist, user either closed the figure or
            # is using IPython, in this case we need a new figure.

            fig = plt.figure(figsize=figsize)
            updatefig = False
        else:
            # Set current figure to given input
            fig = plt.figure(fig.number)
            updatefig = True

            if values.ndim > 1 and not update_in_place:
                # If the figure is larger than 1d, we can clear it since we
                # dont reuse anything. Keeping it causes performance problems.
                fig.clf()
    else:
        fig = plt.figure(figsize=figsize)
        updatefig = False

    if values_are_complex:
        # Real
        if len(fig.axes) == 0:
            # Create new axis if needed
            sub_re = plt.subplot(arrange_subplots[0], **sub_kwargs)
            sub_re.set_title('Real part')
            sub_re.set_xlabel(axis_labels[0], fontsize=axis_fontsize)
            if values.ndim == 2:
                sub_re.set_ylabel(axis_labels[1], fontsize=axis_fontsize)
            else:
                sub_re.set_ylabel('value')
        else:
            sub_re = fig.axes[0]

        display_re = getattr(sub_re, method)
        csub_re = display_re(*args_re, **dsp_kwargs)

        # Axis ticks
        if method == 'imshow' and not grid.is_uniform:
            (xpts, xlabels), (ypts, ylabels) = _axes_info(grid)
            plt.xticks(xpts, xlabels)
            plt.yticks(ypts, ylabels)

        if method == 'imshow' and len(fig.axes) < 2:
            # Create colorbar if none seems to exist

            # Use clim from kwargs if given
            if 'clim' not in kwargs:
                minval_re, maxval_re = _safe_minmax(values.real)
            else:
                minval_re, maxval_re = kwargs['clim']

            ticks_re = _colorbar_ticks(minval_re, maxval_re)
            fmt_re = _colorbar_format(minval_re, maxval_re)

            plt.colorbar(csub_re, orientation='horizontal',
                         ticks=ticks_re, format=fmt_re)

        # Imaginary
        if len(fig.axes) < 3:
            sub_im = plt.subplot(arrange_subplots[1], **sub_kwargs)
            sub_im.set_title('Imaginary part')
            sub_im.set_xlabel(axis_labels[0], fontsize=axis_fontsize)
            if values.ndim == 2:
                sub_im.set_ylabel(axis_labels[1], fontsize=axis_fontsize)
            else:
                sub_im.set_ylabel('value')
        else:
            sub_im = fig.axes[2]

        display_im = getattr(sub_im, method)
        csub_im = display_im(*args_im, **dsp_kwargs)

        # Axis ticks
        if method == 'imshow' and not grid.is_uniform:
            (xpts, xlabels), (ypts, ylabels) = _axes_info(grid)
            plt.xticks(xpts, xlabels)
            plt.yticks(ypts, ylabels)

        if method == 'imshow' and len(fig.axes) < 4:
            # Create colorbar if none seems to exist

            # Use clim from kwargs if given
            if 'clim' not in kwargs:
                minval_im, maxval_im = _safe_minmax(values.imag)
            else:
                minval_im, maxval_im = kwargs['clim']

            ticks_im = _colorbar_ticks(minval_im, maxval_im)
            fmt_im = _colorbar_format(minval_im, maxval_im)

            plt.colorbar(csub_im, orientation='horizontal',
                         ticks=ticks_im, format=fmt_im)

    else:
        if len(fig.axes) == 0:
            # Create new axis object if needed
            sub = plt.subplot(111, **sub_kwargs)
            sub.set_xlabel(axis_labels[0], fontsize=axis_fontsize)
            if values.ndim == 2:
                sub.set_ylabel(axis_labels[1], fontsize=axis_fontsize)
            else:
                sub.set_ylabel('value')
            try:
                # For 3d plots
                sub.set_zlabel('z')
            except AttributeError:
                pass
        else:
            sub = fig.axes[0]

        if update_in_place:
            import matplotlib as mpl
            imgs = [obj for obj in sub.get_children()
                    if isinstance(obj, mpl.image.AxesImage)]
            if len(imgs) > 0 and updatefig:
                imgs[0].set_data(args_re[0])
                csub = imgs[0]

                # Update min-max
                if 'clim' not in kwargs:
                    minval, maxval = _safe_minmax(values)
                else:
                    minval, maxval = kwargs['clim']

                csub.set_clim(minval, maxval)
            else:
                display = getattr(sub, method)
                csub = display(*args_re, **dsp_kwargs)
        else:
            display = getattr(sub, method)
            csub = display(*args_re, **dsp_kwargs)

        # Axis ticks
        if method == 'imshow' and not grid.is_uniform:
            (xpts, xlabels), (ypts, ylabels) = _axes_info(grid)
            plt.xticks(xpts, xlabels)
            plt.yticks(ypts, ylabels)

        if method == 'imshow' and colorbar:
            # Add colorbar
            # Use clim from kwargs if given
            if 'clim' not in kwargs:
                minval, maxval = _safe_minmax(values)
            else:
                minval, maxval = kwargs['clim']

            ticks = _colorbar_ticks(minval, maxval)
            fmt = _colorbar_format(minval, maxval)
            if len(fig.axes) < 2:
                # Create colorbar if none seems to exist
                plt.colorbar(mappable=csub, ticks=ticks, format=fmt)
            elif update_in_place:
                # If it exists and we should update it
                csub.set_clim(minval, maxval)
                csub.colorbar.set_ticks(ticks)
                if '%' not in fmt:
                    labels = [fmt] * len(ticks)
                else:
                    labels = [fmt % t for t in ticks]
                csub.colorbar.set_ticklabels(labels)
                fig.canvas.draw_idle()

    # Set title of window
    if title is not None:
        if not values_are_complex:
            # Do not overwrite title for complex values
            plt.title(title)
        fig.canvas.manager.set_window_title(title)

    # Fixes overlapping stuff at the expense of potentially squashed subplots
    if not update_in_place:
        fig.tight_layout()

    if updatefig or plt.isinteractive():
        # If we are running in interactive mode, we can always show the fig
        # This causes an artifact, where users of `CallbackShow` without
        # interactive mode only shows the figure after the second iteration.
        plt.show(block=False)
        if not update_in_place:
            plt.draw()
            warning_free_pause()
        else:
            try:
                sub.draw_artist(csub)
                fig.canvas.blit(fig.bbox)
                fig.canvas.update()
                fig.canvas.flush_events()
            except AttributeError:
                plt.draw()
                warning_free_pause()

    if force_show:
        plt.show()

    if saveto is not None:
        fig.savefig(saveto)
    return fig


if __name__ == '__main__':
    run_doctests()



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import division

import numpy as np
import pytest

import odl
from odl.core.discr.grid import sparse_meshgrid
from odl.core.set.domain import IntervalProd
from odl.core.util.testutils import all_equal


def random_point(set_):
    if isinstance(set_, IntervalProd):
        return (set_.min_pt +
                np.random.rand(set_.ndim) * (set_.max_pt - set_.min_pt))
    else:
        raise NotImplementedError("unknown type")


def test_init():
    IntervalProd(1, 2)
    IntervalProd(-np.inf, 2)
    IntervalProd(0, np.inf)
    IntervalProd([1], [2])
    IntervalProd((1,), (2,))

    IntervalProd([1, 2, 3], [4, 5, 6])
    IntervalProd((1, 2, 3), (4, 5, 6))

    IntervalProd((1, 2, 3), (1, 2, 3))

    with pytest.raises(ValueError):
        IntervalProd(2, 1)

    with pytest.raises(ValueError):
        IntervalProd((1, 2, 3), (1, 2, 0))


def test_min_pt():
    set_ = IntervalProd(1, 2)
    assert set_.min_pt == 1

    set_ = IntervalProd(-np.inf, 0)
    assert set_.min_pt == -np.inf

    set_ = IntervalProd([1], [2])
    assert set_.min_pt == 1

    set_ = IntervalProd([1, 2, 3], [5, 6, 7])
    assert all_equal(set_.min_pt, [1, 2, 3])


def test_max_pt():
    set_ = IntervalProd(1, 2)
    assert set_.max_pt == 2

    set_ = IntervalProd(0, np.inf)
    assert set_.max_pt == np.inf

    set_ = IntervalProd([1], [2])
    assert set_.max_pt == 2

    set_ = IntervalProd([1, 2, 3], [5, 6, 7])
    assert all_equal(set_.max_pt, [5, 6, 7])


def test_ndim():
    set_ = IntervalProd(1, 2)
    assert set_.ndim == 1

    set_ = IntervalProd(1, 1)
    assert set_.ndim == 1

    set_ = IntervalProd(0, np.inf)
    assert set_.ndim == 1

    set_ = IntervalProd([1], [2])
    assert set_.ndim == 1

    set_ = IntervalProd([1, 2, 3], [5, 6, 7])
    assert set_.ndim == 3

    set_ = IntervalProd([1, 2, 3], [1, 6, 7])
    assert set_.ndim == 3


def test_true_ndim():
    set_ = IntervalProd(1, 2)
    assert set_.true_ndim == 1

    set_ = IntervalProd(1, 1)
    assert set_.true_ndim == 0

    set_ = IntervalProd(0, np.inf)
    assert set_.true_ndim == 1

    set_ = IntervalProd([1], [2])
    assert set_.true_ndim == 1

    set_ = IntervalProd([1, 2, 3], [5, 6, 7])
    assert set_.true_ndim == 3

    set_ = IntervalProd([1, 2, 3], [1, 6, 7])
    assert set_.true_ndim == 2


def test_extent():
    set_ = IntervalProd(1, 2)
    assert set_.extent == 1

    set_ = IntervalProd(1, 1)
    assert set_.extent == 0

    set_ = IntervalProd(0, np.inf)
    assert set_.extent == np.inf

    set_ = IntervalProd(-np.inf, 0)
    assert set_.extent == np.inf

    set_ = IntervalProd(-np.inf, np.inf)
    assert set_.extent == np.inf

    set_ = IntervalProd([1, 2, 3], [5, 6, 7])
    assert list(set_.extent) == [4, 4, 4]


def test_volume():
    set_ = IntervalProd(1, 2)
    assert set_.volume == 2 - 1

    set_ = IntervalProd(0, np.inf)
    assert set_.volume == np.inf

    set_ = IntervalProd([1, 2, 3], [5, 6, 7])
    assert set_.volume == pytest.approx((5 - 1) * (6 - 2) * (7 - 3))


def test_mid_pt():
    set_ = IntervalProd(1, 2)
    assert set_.mid_pt == 1.5

    set_ = IntervalProd(0, np.inf)
    assert set_.mid_pt == np.inf

    set_ = IntervalProd([1, 2, 3], [5, 6, 7])
    assert all_equal(set_.mid_pt, [3, 4, 5])


def test_element():
    set_ = IntervalProd(1, 2)
    assert set_.element() in set_

    set_ = IntervalProd(0, np.inf)
    assert set_.element() in set_

    set_ = IntervalProd([1, 2, 3], [5, 6, 7])
    assert set_.element() in set_


def _test_eq(x, y):
    """Test equality of x and y."""
    assert x == y
    assert not x != y
    assert hash(x) == hash(y)


def _test_neq(x, y):
    """Test non-equality of x and y."""
    assert x != y
    assert not x == y
    assert hash(x) != hash(y)


def test_equals():
    """Test equality check of IntervalProd."""
    interval1 = IntervalProd(1, 2)
    interval2 = IntervalProd(1, 2)
    interval3 = IntervalProd([1], [2])
    interval4 = IntervalProd(2, 3)
    rectangle1 = IntervalProd([1, 2], [2, 3])
    rectangle2 = IntervalProd((1, 2), (2, 3))
    rectangle3 = IntervalProd([0, 2], [2, 3])

    _test_eq(interval1, interval1)
    _test_eq(interval1, interval2)
    _test_eq(interval1, interval3)
    _test_neq(interval1, interval4)
    _test_neq(interval1, rectangle1)
    _test_eq(rectangle1, rectangle1)
    _test_eq(rectangle2, rectangle2)
    _test_eq(rectangle1, rectangle2)
    _test_neq(rectangle1, rectangle3)

    r1_1 = IntervalProd(-np.inf, np.inf)
    r1_2 = IntervalProd(-np.inf, np.inf)
    positive_reals = IntervalProd(0, np.inf)
    _test_eq(r1_1, r1_1)
    _test_eq(r1_1, r1_2)
    _test_eq(positive_reals, positive_reals)
    _test_neq(positive_reals, r1_1)

    # non interval
    for non_interval in [1, 1j, np.array([1, 2])]:
        assert interval1 != non_interval
        assert rectangle1 != non_interval


def test_contains():
    set_ = IntervalProd(1, 2)

    assert 1 in set_
    assert 2 in set_
    assert 1.5 in set_
    assert 3 not in set_
    assert 'string' not in set_
    assert [1, 2] not in set_
    assert np.nan not in set_

    positive_reals = IntervalProd(0, np.inf)
    assert 1 in positive_reals
    assert np.inf in positive_reals
    assert -np.inf not in positive_reals
    assert -1 not in positive_reals


def test_contains_set():
    set_ = IntervalProd(1, 2)

    for sub_set in [np.array([1, 1.1, 1.2, 1.3, 1.4]),
                    IntervalProd(1.2, 2),
                    IntervalProd(1, 1.5),
                    IntervalProd(1.2, 1.2)]:
        assert set_.contains_set(sub_set)

    with np.errstate(invalid="ignore"):
        for non_sub_set in [np.array([0, 1, 1.1, 1.2, 1.3, 1.4]),
                            np.array([np.nan, 1.1, 1.3]),
                            IntervalProd(1.2, 3),
                            IntervalProd(0, 1.5),
                            IntervalProd(3, 4)]:
            assert not set_.contains_set(non_sub_set)

    for non_set in [1,
                    [1, 2],
                    {'hello': 1.0}]:
        with pytest.raises(AttributeError):
            set_.contains_set(non_set)


def test_contains_all():
    # 1d
    intvp = IntervalProd(1, 2)

    arr_in1 = np.array([1.0, 1.6, 1.3, 2.0])
    arr_in2 = np.array([[1.0, 1.6, 1.3, 2.0]])
    mesh_in = sparse_meshgrid([1.0, 1.7, 1.9])
    arr_not_in1 = np.array([1.0, 1.6, 1.3, 2.0, 0.8])
    arr_not_in2 = np.array([[1.0, 1.6, 2.0], [1.0, 1.5, 1.6]])
    mesh_not_in = sparse_meshgrid([-1.0, 1.7, 1.9])

    assert intvp.contains_all(arr_in1)
    assert intvp.contains_all(arr_in2)
    assert intvp.contains_all(mesh_in)
    assert not intvp.contains_all(arr_not_in1)
    assert not intvp.contains_all(arr_not_in2)
    assert not intvp.contains_all(mesh_not_in)

    # 2d
    intvp = IntervalProd([0, 0], [1, 2])
    arr_in = np.array([[0.5, 1.9],
                       [0.0, 1.0],
                       [1.0, 0.1]]).T
    mesh_in = sparse_meshgrid([0, 0.1, 0.6, 0.7, 1.0], [0])
    arr_not_in = np.array([[0.5, 1.9],
                           [1.1, 1.0],
                           [1.0, 0.1]]).T
    mesh_not_in = sparse_meshgrid([0, 0.1, 0.6, 0.7, 1.0], [0, -1])

    assert intvp.contains_all(arr_in)
    assert intvp.contains_all(mesh_in)
    assert not intvp.contains_all(arr_not_in)
    assert not intvp.contains_all(mesh_not_in)


def test_insert():
    intvp1 = IntervalProd([0, 0], [1, 2])
    intvp2 = IntervalProd(1, 3)

    intvp = intvp1.insert(0, intvp2)
    true_min_pt = [1, 0, 0]
    true_max_pt = [3, 1, 2]
    assert intvp == IntervalProd(true_min_pt, true_max_pt)

    intvp = intvp1.insert(1, intvp2)
    true_min_pt = [0, 1, 0]
    true_max_pt = [1, 3, 2]
    assert intvp == IntervalProd(true_min_pt, true_max_pt)

    intvp = intvp1.insert(2, intvp2)
    true_min_pt = [0, 0, 1]
    true_max_pt = [1, 2, 3]
    assert intvp == IntervalProd(true_min_pt, true_max_pt)

    intvp = intvp1.insert(-1, intvp2)  # same as 1
    true_min_pt = [0, 1, 0]
    true_max_pt = [1, 3, 2]
    assert intvp == IntervalProd(true_min_pt, true_max_pt)

    with pytest.raises(IndexError):
        intvp1.insert(3, intvp2)
    with pytest.raises(IndexError):
        intvp1.insert(-4, intvp2)


def test_dist():
    set_ = IntervalProd(1, 2)

    for interior in [1.0, 1.1, 2.0]:
        assert set_.dist(interior) == 0.0

    for exterior in [0.0, 2.0, np.inf]:
        assert set_.dist(exterior) == min(abs(set_.min_pt - exterior),
                                          abs(exterior - set_.max_pt))

    assert set_.dist(np.nan) == np.inf


# Set arithmetic
def test_pos():
    interv = IntervalProd(1, 2)
    assert +interv == interv

    interv = IntervalProd([1, 2, 3, 4, 5], [2, 3, 4, 5, 6])
    assert +interv == interv


def test_neg():
    interv = IntervalProd(1, 2)
    assert -interv == IntervalProd(-2, -1)

    interv = IntervalProd([1, 2, 3, 4, 5], [2, 3, 4, 5, 6])
    assert -interv == IntervalProd([-2, -3, -4, -5, -6],
                                   [-1, -2, -3, -4, -5])


def test_add():
    interv1 = IntervalProd(1, 2)
    interv2 = IntervalProd(3, 4)

    assert interv1 + 2.0 == IntervalProd(3, 4)
    assert interv1 + interv2 == IntervalProd(4, 6)

    interv1 = IntervalProd([1, 2, 3, 4, 5], [2, 3, 4, 5, 6])
    interv2 = IntervalProd([1, 2, 3, 4, 5], [2, 3, 4, 5, 6])

    assert interv1 + interv2 == IntervalProd([2, 4, 6, 8, 10],
                                             [4, 6, 8, 10, 12])


def test_sub():
    interv1 = IntervalProd(1, 2)
    interv2 = IntervalProd(3, 4)

    assert interv1 - 2.0 == IntervalProd(-1, 0)
    assert interv1 - interv2 == IntervalProd(-3, -1)

    interv1 = IntervalProd([1, 2, 3, 4, 5], [2, 3, 4, 5, 6])
    interv2 = IntervalProd([1, 2, 3, 4, 5], [2, 3, 4, 5, 6])

    assert interv1 - interv2 == IntervalProd([-1, -1, -1, -1, -1],
                                             [1, 1, 1, 1, 1])


def test_mul():
    interv1 = IntervalProd(1, 2)
    interv2 = IntervalProd(3, 4)

    assert interv1 * 2.0 == IntervalProd(2, 4)
    assert interv1 * interv2 == IntervalProd(3, 8)

    interv1 = IntervalProd([1, 2, 3, 4, 5], [2, 3, 4, 5, 6])
    interv2 = IntervalProd([1, 2, 3, 4, 5], [2, 3, 4, 5, 6])

    assert interv1 * interv2 == IntervalProd([1, 4, 9, 16, 25],
                                             [4, 9, 16, 25, 36])


def test_div():
    interv1 = IntervalProd(1, 2)
    interv2 = IntervalProd(3, 4)

    assert interv1 / 2.0 == IntervalProd(1 / 2.0, 2 / 2.0)
    assert 2.0 / interv1 == IntervalProd(2 / 2.0, 2 / 1.0)
    assert interv1 / interv2 == IntervalProd(1 / 4.0, 2.0 / 3.0)

    interv_with_zero = IntervalProd(-1, 1)
    with pytest.raises(ValueError):
        interv1 / interv_with_zero

    interv1 = IntervalProd([1, 2, 3, 4, 5], [2, 3, 4, 5, 6])
    interv2 = IntervalProd([1, 2, 3, 4, 5], [2, 3, 4, 5, 6])
    quotient = IntervalProd([1 / 2., 2 / 3., 3 / 4., 4 / 5., 5 / 6.],
                            [2 / 1., 3 / 2., 4 / 3., 5 / 4., 6 / 5.])

    assert (interv1 / interv2).approx_equals(quotient, atol=1e-10)


def test_interval_length():
    set_ = IntervalProd(1, 2)
    assert set_.length == set_.volume
    assert set_.length == 1


def test_rectangle_area():
    set_ = IntervalProd([1, 2], [3, 4])
    assert set_.area == set_.volume
    assert set_.area == (3 - 1) * (4 - 2)


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Standardized tests for ``LinearSpace``."""

from __future__ import absolute_import, division, print_function

from builtins import object
from copy import copy, deepcopy

from odl.core.diagnostics.examples import samples
from odl.core.set import Field
from odl.core.util.testutils import fail_counter

__all__ = ('SpaceTest',)


def _approx_equal(x, y, eps):
    """Test if elements ``x`` and ``y`` are approximately equal.

    ``eps`` is a given absolute tolerance.
    """
    if x.space != y.space:
        return False

    if x is y:
        return True

    try:
        return x.dist(y) <= eps
    except NotImplementedError:
        try:
            return x == y
        except NotImplementedError:
            return False


class SpaceTest(object):

    """Automated tests for `LinearSpace` instances.

    This class allows users to automatically test various features of a
    `LinearSpace` such as linearity and vector space operations.
    """

    def __init__(self, space, verbose=True, tol=1e-5):
        """Initialize a new instance.

        Parameters
        ----------
        space : `LinearSpace`
            Space that should be tested.
        verbose : bool, optional
            If ``True``, print additional info text.
        tol : float, optional
            Tolerance parameter used as a base for the actual tolerance
            in the tests. Depending on the expected accuracy, the actual
            tolerance used in a test can be a factor times this number.
        """
        self.space = space
        self.verbose = bool(verbose)
        self.tol = float(tol)

    def log(self, message):
        """Print message if ``self.verbose == True``."""
        if self.verbose:
            print(message)

    def element_method(self):
        """Verify `LinearSpace.element`."""
        with fail_counter(
            test_name='Verifying element method', logger=self.log
        ) as counter:
            try:
                elem = self.space.element()
            except NotImplementedError:
                counter.fail('*** element failed ***')
                return

            if elem not in self.space:
                counter.fail('*** space.element() not in space ***')

    def field(self):
        """Verify `LinearSpace.field`."""
        with fail_counter(
            test_name='Verifying field property', logger=self.log
        ) as counter:
            try:
                field = self.space.field
            except NotImplementedError:
                counter.fail('*** field failed ***')
                return

            if not isinstance(field, Field):
                counter.fail('*** space.field not a `Field` ***')
                return

            try:
                zero = field.element(0)
            except NotImplementedError:
                counter.fail('*** field.element(0) failed ***')
                zero = None

            if zero is not None and zero != 0:
                counter.fail('*** field.element(0) != 0 ***')

            if zero is not None and zero != 0.0:
                counter.fail('*** field.element(0) != 0.0 ***')

            try:
                one = field.element(1)
            except NotImplementedError:
                counter.fail('*** field.element(1) failed ***')
                one = None

            if one is not None and one != 1:
                counter.fail('*** field.element(1) != 1 ***')

            if one is not None and one != 1.0:
                counter.fail('*** field.element(1) != 1.0 ***')

            try:
                minus_one = field.element(-1)
            except NotImplementedError:
                counter.fail('field.element(-1) failed')
                minus_one = None

            if minus_one is not None and minus_one != -1:
                counter.fail('field.element(-1) != -1')

            if minus_one is not None and minus_one != -1.0:
                counter.fail('field.element(-1) != -1.0')

    def _associativity_of_addition(self):
        """Verify addition associativity."""
        with fail_counter(
            test_name='Verifying associativity of addition',
            err_msg='error = dist(x + (y + z), (x + y) + z)',
            logger=self.log
        ) as counter:

            for [n_x, x], [n_y, y], [n_z, z] in samples(self.space,
                                                        self.space,
                                                        self.space):
                correct = _approx_equal(x + (y + z), (x + y) + z, self.tol)
                if not correct:
                    counter.fail('failed with x={:25s} y={:25s} z={:25s}'
                                 ''.format(n_x, n_y, n_z))

    def _commutativity_of_addition(self):
        """Verify addition commutativity."""
        with fail_counter(
            test_name='Verifying commutativity of addition',
            err_msg='error = dist(x + y, y + x)',
            logger=self.log
        ) as counter:

            for [n_x, x], [n_y, y] in samples(self.space, self.space):
                correct = _approx_equal(x + y, y + x, self.tol)
                if not correct:
                    counter.fail('failed with x={:25s} y={:25s}'
                                 ''.format(n_x, n_y))

    def _identity_of_addition(self):
        """Verify additive neutral element ('zero')."""
        try:
            zero = self.space.zero()
        except (AttributeError, NotImplementedError):
            print('*** SPACE HAS NO ZERO VECTOR ***')
            return

        with fail_counter(
            test_name='Verifying identity element of addition',
            err_msg='error = dist(x + 0, x)',
            logger=self.log
        ) as counter:

            for [n_x, x] in samples(self.space):
                correct = _approx_equal(x + zero, x, self.tol)
                if not correct:
                    counter.fail('failed with x={:25s}'.format(n_x))

    def _inverse_element_of_addition(self):
        """Verify additive inverse."""
        try:
            zero = self.space.zero()
        except (AttributeError, NotImplementedError):
            print('*** SPACE HAS NO ZERO VECTOR ***')
            return

        with fail_counter(
            test_name='Verifying inverse element of addition',
            err_msg='error = dist(x + (-x), 0)',
            logger=self.log
        ) as counter:

            for [n_x, x] in samples(self.space):
                correct = _approx_equal(x + (-x), zero, self.tol)
                if not correct:
                    counter.fail('failed with x={:25s}'.format(n_x))

    def _commutativity_of_scalar_mult(self):
        """Verify scalar multiplication commutativity."""
        with fail_counter(
            test_name='Verifying commutativity of scalar multiplication',
            err_msg='error = dist(a * (b * x), (a * b) * x)',
            logger=self.log
        ) as counter:

            for [n_x, x], [_, a], [_, b] in samples(self.space,
                                                    self.space.field,
                                                    self.space.field):
                correct = _approx_equal(a * (b * x), (a * b) * x, self.tol)
                if not correct:
                    counter.fail('failed with x={:25s}, a={}, b={}'
                                 ''.format(n_x, a, b))

    def _identity_of_mult(self):
        """Verify multiplicative neutral element ('one')."""
        with fail_counter(
            test_name='Verifying identity element of multiplication',
            err_msg='error = dist(1 * x, x)',
            logger=self.log
        ) as counter:

            for [n_x, x] in samples(self.space):
                correct = _approx_equal(1 * x, x, self.tol)
                if not correct:
                    counter.fail('failed with x={:25s}'.format(n_x))

    def _distributivity_of_mult_vector(self):
        """Verify scalar multiplication distributivity wrt vector addition."""
        with fail_counter(
            test_name='Verifying distributivity of scalar multiplication '
                      'under vector addition',
            err_msg='error = dist(a * (x + y), a * x + a * y)',
            logger=self.log
        ) as counter:

            for [n_x, x], [n_y, y], [_, a] in samples(self.space,
                                                      self.space,
                                                      self.space.field):
                correct = _approx_equal(a * (x + y), a * x + a * y, self.tol)
                if not correct:
                    counter.fail('failed with x={:25s}, y={:25s}, a={}'
                                 ''.format(n_x, n_y, a))

    def _distributivity_of_mult_scalar(self):
        """Verify scalar multiplication distributivity wrt scalar addition."""
        with fail_counter(
            test_name='Verifying distributivity of scalar multiplication '
                      'under scalar addition',
            err_msg='error = dist((a + b) * x, a * x + b * x)',
            logger=self.log
        ) as counter:

            for [n_x, x], [_, a], [_, b] in samples(self.space,
                                                    self.space.field,
                                                    self.space.field):
                correct = _approx_equal((a + b) * x, a * x + b * x, self.tol)
                if not correct:
                    counter.fail('failed with x={:25s}, a={}, b={}'
                                 ''.format(n_x, a, b))

    def _subtraction(self):
        """Verify element subtraction as addition of additive inverse."""
        with fail_counter(
            test_name='Verifying element subtraction',
            err_msg='error = dist(x - y, x + (-1 * y))',
            logger=self.log
        ) as counter:

            for [n_x, x], [n_y, y] in samples(self.space, self.space):
                correct = (_approx_equal(x - y, x + (-1 * y), self.tol) and
                           _approx_equal(x - y, x + (-y), self.tol))
                if not correct:
                    counter.fail('failed with x={:25s}, y={:25s}'
                                 ''.format(n_x, n_y))

    def _division(self):
        """Verify scalar division as multiplication with mult. inverse."""
        with fail_counter(
            test_name='Verifying scalar division',
            err_msg='error = dist(x / a, x * (1/a))',
            logger=self.log
        ) as counter:

            for [n_x, x], [_, a] in samples(self.space, self.space.field):
                if a != 0:
                    correct = _approx_equal(x / a, x * (1.0 / a), self.tol)
                    if not correct:
                        counter.fail('failed with x={:25s}, a={}'
                                     ''.format(n_x, a))

    def _lincomb_aliased(self):
        """Verify several scenarios of aliased linear combination."""
        with fail_counter(
            test_name='Verifying linear combination with aliased input',
            err_msg='error = dist(aliased, non-aliased)',
            logger=self.log
        ) as counter:

            for [n_x, x_in], [n_y, y] in samples(self.space, self.space):
                x = x_in.copy()
                x.lincomb(1, x, 1, y)
                correct = _approx_equal(x, x_in + y, self.tol)
                if not correct:
                    counter.fail('failed with x.lincomb(1, x, 1, y),'
                                 'x={:25s} y={:25s} '
                                 ''.format(n_x, n_y))

                x = x_in.copy()
                x.lincomb(1, x, 1, x)
                correct = _approx_equal(x, x_in + x_in, self.tol)
                if not correct:
                    counter.fail('failed with x.lincomb(1, x, 1, x),'
                                 'x={:25s} '
                                 ''.format(n_x))

    def _lincomb(self):
        """Verify linear combination."""
        self.log('\nTesting lincomb')
        self._lincomb_aliased()

    def linearity(self):
        """Verify the linear space properties by examples.

        These properties include things such as associativity

            ``x + y = y + x``

        and identity of the ``LinearSpace.zero`` element

            ``x + 0 = x``

        References
        ----------
        Wikipedia article on `Vector space`_.

        .. _Vector space: https://en.wikipedia.org/wiki/Vector_space
        """
        self.log('\n== Verifying linear space properties ==\n')
        self._associativity_of_addition()
        self._commutativity_of_addition()
        self._identity_of_addition()
        self._inverse_element_of_addition()
        self._commutativity_of_scalar_mult()
        self._identity_of_mult()
        self._distributivity_of_mult_vector()
        self._distributivity_of_mult_scalar()
        self._subtraction()
        self._division()
        self._lincomb()

    def _inner_linear_scalar(self):
        """Verify homogeneity of the inner product in the first argument."""
        with fail_counter(
            test_name='Verifying homogeneity of the inner product in the '
                      'first argument',
            err_msg='error = |<a*x, y> - a*<x, y>|',
            logger=self.log
        ) as counter:

            for [n_x, x], [n_y, y], [_, a] in samples(self.space,
                                                      self.space,
                                                      self.space.field):
                error = abs((a * x).inner(y) - a * x.inner(y))
                if error > self.tol:
                    counter.fail('x={:25s}, y={:25s}, a={}: error={}'
                                 ''.format(n_x, n_y, a, error))

    def _inner_conjugate_symmetry(self):
        """Verify conjugate symmetry of the inner product."""
        with fail_counter(
            test_name='Verifying conjugate symmetry of the inner product',
            err_msg='error = |<x, y> - <y, x>.conj()|',
            logger=self.log
        ) as counter:

            for [n_x, x], [n_y, y] in samples(self.space, self.space):
                error = abs((x).inner(y) - y.inner(x).conjugate())
                if error > self.tol:
                    counter.fail('x={:25s}, y={:25s}: error={}'
                                 ''.format(n_x, n_y, error))

    def _inner_linear_sum(self):
        """Verify distributivity of the inner product in the first argument."""
        with fail_counter(
            test_name='Verifying distributivity of the inner product'
                      'in the first argument',
            err_msg='error = |<x+y, z> - (<x, z> + <y, z>)|',
            logger=self.log
        ) as counter:

            for [n_x, x], [n_y, y], [n_z, z] in samples(self.space,
                                                        self.space,
                                                        self.space):
                error = abs((x + y).inner(z) - (x.inner(z) + y.inner(z)))
                if error > self.tol:
                    counter.fail('x={:25s}, y={:25s}, z={:25s}: error={}'
                                 ''.format(n_x, n_y, n_z, error))

    def _inner_positive(self):
        """Verify positive definiteness of the inner product."""
        with fail_counter(
            test_name='Verifying positive definiteness of the inner '
                      'product',
            logger=self.log
        ) as counter:

            for [n_x, x] in samples(self.space):
                inner = x.inner(x)

                if abs(inner.imag) > self.tol:
                    counter.fail('<x, x>.imag != 0, x={:25s}, <x, x>.imag = {}'
                                 ''.format(n_x, inner.imag))

                if n_x == 'Zero' and inner.real != 0:
                    counter.fail('<0, 0> != 0.0, x={:25s}: <0, 0>={}'
                                 ''.format(n_x, inner))

                elif n_x != 'Zero' and inner.real <= 0:
                    counter.fail('<x, x> <= 0,   x={:25s}: <x, x>={}'
                                 ''.format(n_x, inner))

    def inner(self):
        """Verify ``LinearSpace.inner``.

        The inner product is checked for the following properties:

        - conjugate symmetry:

            ``<x, y> = <y, x>^*`` (^* complex conjugate)

        - linearity:

            ``<a * x, y> = a * <x, y>``

            ``<x + y, z> = <x, z> + <y, z>``

        - positive definiteness:

            ``<x, x> >= 0``

        References
        ----------
        Wikipedia article on `inner product`_.

        .. _inner product: https://en.wikipedia.org/wiki/Inner_product_space
        """
        self.log('\n== Verifying inner product ==\n')

        try:
            zero = self.space.zero()
            zero.inner(zero)
        except NotImplementedError:
            self.log('Space has no inner product')
            return

        self._inner_conjugate_symmetry()
        self._inner_linear_scalar()
        self._inner_linear_sum()
        self._inner_positive()

    def _norm_positive(self):
        """Verify positive definiteness of the norm."""
        with fail_counter(
            test_name='Verifying positive definiteness of the norm',
            logger=self.log
        ) as counter:

            for [n_x, x] in samples(self.space):
                norm = x.norm()

                if n_x == 'Zero' and norm != 0:
                    counter.fail('||0|| != 0.0, x={:25s}: ||x||={}'
                                 ''.format(n_x, norm))

                elif n_x != 'Zero' and norm <= 0:
                    counter.fail('||x|| <= 0,   x={:25s}: ||x||={}'
                                 ''.format(n_x, norm))

    def _norm_subadditive(self):
        """Verify subadditivity of the norm."""
        with fail_counter(
            test_name='Verifying sub-additivity of the norm',
            err_msg='error = max(||x+y|| - (||x|| + ||y||), 0)',
            logger=self.log
        ) as counter:

            for [n_x, x], [n_y, y] in samples(self.space, self.space):
                norm_x = x.norm()
                norm_y = y.norm()
                norm_xy = (x + y).norm()

                error = norm_xy - norm_x - norm_y

                if error > 0:
                    counter.fail('x={:25s} y={:25s}: error={}'
                                 ''.format(n_x, n_y, error))

    def _norm_homogeneity(self):
        """Verify positive homogeneity of the norm."""
        with fail_counter(
            test_name='Verifying positive homogeneity of the norm',
            err_msg='error = | ||a*x|| - |a|*||x|| |',
            logger=self.log
        ) as counter:

            for [n_x, x], [_, a] in samples(self.space, self.space.field):
                error = abs((a * x).norm() - abs(a) * x.norm())
                if error > self.tol:
                    counter.fail('x={:25s} a={}: error={}'
                                 ''.format(n_x, a, error))

    def _norm_inner_compatible(self):
        """Verify compatibility of norm and inner product."""
        try:
            zero = self.space.zero()
            zero.inner(zero)
        except NotImplementedError:
            self.log('Space has no inner product')
            return

        with fail_counter(
            test_name='Verifying compatibility of norm and inner product',
            err_msg='error = | ||x||^2 - <x, x> |',
            logger=self.log
        ) as counter:

            for [n_x, x] in samples(self.space):
                error = abs(x.norm() ** 2 - x.inner(x))

                if error > self.tol:
                    counter.fail('x={:25s}: error={}'
                                 ''.format(n_x, error))

    def norm(self):
        """Verify ``LinearSpace.norm``.

        The norm is checked for the following properties:

        - linearity:

            ``||a * x|| = |a| * ||x||``

        - subadditivity:

            ``||x + y|| <= ||x|| + ||y||``

        - positive homogeneity:

            ``||a * x|| = |a| * ||x||``

        - positive definiteness:

            ``||x|| >= 0``

            ``||x|| = 0`` iff ``x = 0``

        - compatibility with the inner product (if available):

            ``||x||^2 = <x, x>``

        References
        ----------
        Wikipedia article on norm_.

        .. _norm: https://en.wikipedia.org/wiki/Norm_(mathematics)
        """
        self.log('\n== Verifying norm ==\n')

        try:
            self.space.zero().norm()
        except NotImplementedError:
            self.log('Space has no norm')
            return

        self._norm_positive()
        self._norm_subadditive()
        self._norm_homogeneity()
        self._norm_inner_compatible()

    def _dist_positivity(self):
        """Verify nonnegativity of the distance."""

        with fail_counter(
            test_name='Verifying nonnegativity of the distance',
            logger=self.log
        ) as counter:

            for [n_x, x], [n_y, y] in samples(self.space, self.space):
                dist = x.dist(y)

                if n_x == n_y and dist != 0:
                    counter.fail('d(x, x) != 0.0, x={:25s}: dist={}'
                                 ''.format(n_x, dist))
                elif n_x != n_y and dist <= 0:
                    counter.fail('d(x, y) <= 0,   x={:25s} y={:25s}: dist={}'
                                 ''.format(n_x, n_y, dist))

    def _dist_symmetric(self):
        """Verify symmetry of the distance."""
        with fail_counter(
            test_name='Verifying symmetry of the distance',
            err_msg='error = |d(x, y) - d(y, x)|',
            logger=self.log
        ) as counter:

            for [n_x, x], [n_y, y] in samples(self.space, self.space):
                dist_1 = x.dist(y)
                dist_2 = y.dist(x)
                error = abs(dist_1 - dist_2)

                if error > self.tol:
                    counter.fail('x={:25s}, y={:25s}: error={}'
                                 ''.format(n_x, n_y, error))

    def _dist_subtransitive(self):
        """Verify sub-transitivity of the distance."""
        with fail_counter(
            test_name='Verifying sub-additivity of the distance',
            err_msg='error = max(d(x,z) - (d(x, y) + d(y, z)), 0)',
            logger=self.log
        ) as counter:

            for [n_x, x], [n_y, y], [n_z, z] in samples(self.space,
                                                        self.space,
                                                        self.space):
                dxz = x.dist(z)
                dxy = x.dist(y)
                dyz = y.dist(z)
                error = dxz - (dxy + dyz)

                if error > self.tol:
                    counter.fail('x={:25s}, y={:25s}, z={:25s}: error={}'
                                 ''.format(n_x, n_y, n_z, error))

    def _dist_norm_compatible(self):
        """Verify compatibility of distance and norm."""
        try:
            self.space.zero().norm()
        except NotImplementedError:
            self.log('Space has no norm')
            return

        with fail_counter(
            test_name='Verifying compatibility of distance and norm',
            err_msg='error = |d(x, y) - ||x-y|| |',
            logger=self.log
        ) as counter:

            for [n_x, x], [n_y, y] in samples(self.space,
                                              self.space):
                error = abs(x.dist(y) - (x - y).norm())

                if error > self.tol:
                    counter.fail('x={:25s}, y={:25s}: error={}'
                                 ''.format(n_x, n_y, error))

    def dist(self):
        """Verify ``LinearSpace.dist``.

        The distance metric is checked for the following properties:

        - positive definiteness:

            ``d(x, y) >= 0``

            ``d(x, y) = 0`` iff ``x = y``

        - symmetry:

            ``d(x, y) = d(y, x)``

        - sub-transitivity:

            ``d(x, z) <= d(x, y) + d(y, z)``

        - compatibility with the norm (if available):

            ``d(x, y) = ||x - y||``

        References
        ----------
        Wikipedia article on metric_

        .. _metric: https://en.wikipedia.org/wiki/Metric_(mathematics)
        """
        self.log('\n== Verifying dist ==\n')

        try:
            zero = self.space.zero()
            self.space.dist(zero, zero)
        except NotImplementedError:
            self.log('Space has no distance metric')
            return

        self._dist_positivity()
        self._dist_symmetric()
        self._dist_subtransitive()
        self._dist_norm_compatible()

    def _multiply_zero(self):
        """Verify that vector multiplication with zero is zero."""
        try:
            zero = self.space.zero()
        except NotImplementedError:
            print('*** SPACE HAS NO ZERO VECTOR ***')
            return

        with fail_counter(
            test_name='Verifying vector multiplication with zero',
            err_msg='error = ||x * 0||',
            logger=self.log
        ) as counter:
            for [n_x, x] in samples(self.space):
                error = (zero * x).norm()

                if error > self.tol:
                    counter.fail('x={:25s},: error={}'
                                 ''.format(n_x, error))

    def _multiply_commutative(self):
        """Verify commutativity of vector multiplication."""
        with fail_counter(
            test_name='Verifying commutativity of vector multiplication',
            err_msg='error = dist(x * y, y * x)',
            logger=self.log
        ) as counter:

            for [n_x, x], [n_y, y], _ in samples(self.space,
                                                 self.space,
                                                 self.space):
                correct = _approx_equal(x * y, y * x, self.tol)
                if not correct:
                    counter.fail('failed with x={:25s} y={:25s}'
                                 ''.format(n_x, n_y))

    def _multiply_associative(self):
        """Verify associativity of vector multiplication."""
        with fail_counter(
            test_name='Verifying associativity of vector multiplication',
            err_msg='error = dist(x * (y * z), (x * y) * z)',
            logger=self.log
        ) as counter:

            for [n_x, x], [n_y, y], [n_z, z] in samples(self.space,
                                                        self.space,
                                                        self.space):
                correct = _approx_equal(x * (y * z), (x * y) * z, self.tol)
                if not correct:
                    counter.fail('failed with x={:25s} y={:25s} z={:25s}'
                                 ''.format(n_x, n_y, n_z))

    def _multiply_distributive_scalar(self):
        """Verify distributivity of scalar multiplication."""
        with fail_counter(
            test_name='Verifying distributivity of vector multiplication '
                      'under scalar multiplication',
            err_msg='error = dist(a * (x + y), a * x + a * y)',
            logger=self.log
        ) as counter:

            for [n_x, x], [n_y, y], [_, a] in samples(self.space,
                                                      self.space,
                                                      self.space.field):
                correct = _approx_equal(a * (x + y), a * x + a * y, self.tol)
                if not correct:
                    counter.fail('failed with x={:25s} y={:25s} a={}'
                                 ''.format(n_x, n_y, a))

    def _multiply_distributive_vector(self):
        """Verify distributivity of vector multiplication."""
        with fail_counter(
            test_name='Verifying distributivity of vector multiplication '
                      'under vector multiplication',
            err_msg='error = dist(x * (y + z), x * y + x * z)',
            logger=self.log
        ) as counter:

            for [n_x, x], [n_y, y], [n_z, z] in samples(self.space,
                                                        self.space,
                                                        self.space):
                correct = _approx_equal(x * (y + z), x * y + x * z, self.tol)
                if not correct:
                    counter.fail('failed with x={:25s} y={:25s} z={:25s}'
                                 ''.format(n_x, n_y, n_z))

    def multiply(self):
        """Verify ``LinearSpace.multiply``.

        The vector multiplication is checked for the following properties:

        - Zero element:

            ``0 * x = 0``

        - Commutativity:

            ``x * y = y * x``

        - Associativity:

            ``x * (y * z) = (x * y) * z``

        - Distributivity:

            ``a * (x + y) = a * x + a * y``

            ``x * (y + z) = x * y + x * z``
        """
        self.log('\n== Verifying multiplication ==\n')

        try:
            zero = self.space.zero()
        except NotImplementedError:
            print('*** SPACE HAS NO ZERO VECTOR ***')
            return

        try:
            self.space.multiply(zero, zero)
        except NotImplementedError:
            self.log('Space has no vector multiplication.')
            return

        self._multiply_zero()
        self._multiply_commutative()
        self._multiply_associative()
        self._multiply_distributive_scalar()
        self._multiply_distributive_vector()

    def equals(self):
        """Verify `LinearSpace.__eq__`."""

        self.log('\n== Verifying __eq__ ==\n')

        if not self.space == self.space:
            print('** space == space failed ***')

        if self.space != self.space:
            print('** not space != space failed***')

        if self.space != copy(self.space):
            print('** space == copy(space) failed***')

        if self.space != deepcopy(self.space):
            print('** space == deepcopy(space) failed***')

        with fail_counter(
            test_name='Verify behavior of `space == obj` when `obj` '
                      'is not a space',
            logger=self.log
        ) as counter:

            for obj in [[1, 2], list(), tuple(), dict(), 5.0]:
                if self.space == obj:
                    counter.fail('space == obj,  with obj={}'
                                 ''.format(obj))

                if not self.space != obj:
                    counter.fail('not space != obj,  with obj={}'
                                 ''.format(obj))

    def contains(self):
        """Verify `LinearSpace.__contains__`."""
        with fail_counter(
            test_name='Verify behavior of `obj in space`',
            logger=self.log
        ) as counter:

            for [n_x, x] in samples(self.space):
                if x not in self.space:
                    counter.fail('x not in space,  with x={}'
                                 ''.format(n_x))

                if x not in self.space:
                    counter.fail('not x in space,  with x={}'
                                 ''.format(n_x))

            for obj in [[1, 2], list(), tuple(), dict(), 5.0]:
                if obj in self.space:
                    counter.fail('obj in space,  with obj={}'
                                 ''.format(obj))

                if not obj not in self.space:
                    counter.fail('not obj not in space,  with obj={}'
                                 ''.format(obj))

    def element_assign(self):
        """Verify `LinearSpaceElement.assign`."""
        with fail_counter(
            test_name='Verify behavior of `LinearSpaceElement.assign`',
            logger=self.log
        ) as counter:

            for [n_x, x], [n_y, y] in samples(self.space,
                                              self.space):
                x.assign(y)
                correct = _approx_equal(x, y, self.tol)
                if not correct:
                    counter.fail('failed with x={:25s} y={:25s}'
                                 ''.format(n_x, n_y))

    def element_copy(self):
        """Verify `LinearSpaceElement.copy`."""
        with fail_counter(
            test_name='Verify behavior of `LinearSpaceElement.copy`',
            logger=self.log
        ) as counter:

            for [n_x, x] in samples(self.space):
                # equal after copy
                y = x.copy()
                correct = _approx_equal(x, y, self.tol)
                if not correct:
                    counter.fail('failed with x={:s5s}'
                                 ''.format(n_x))

                # modify y, x stays the same
                y *= 2.0
                correct = n_x == 'Zero' or not _approx_equal(x, y, self.tol)
                if not correct:
                    counter.fail('modified y, x changed with x={:25s}'
                                 ''.format(n_x))

    def element_set_zero(self):
        """Verify `LinearSpaceElement.set_zero`."""
        try:
            zero = self.space.zero()
        except NotImplementedError:
            print('*** SPACE HAS NO ZERO VECTOR ***')
            return

        with fail_counter(
            test_name='Verify behavior of `LinearSpaceElement.set_zero`',
            logger=self.log
        ) as counter:

            for [n_x, x] in samples(self.space):
                x.set_zero()
                correct = _approx_equal(x, zero, self.tol)
                if not correct:
                    counter.fail('failed with x={:25s}'
                                 ''.format(n_x))

    def element_equals(self):
        """Verify `LinearSpaceElement.__eq__`."""
        try:
            zero = self.space.zero()
        except NotImplementedError:
            print('*** SPACE HAS NO ZERO VECTOR ***')
            return

        try:
            zero == zero
        except NotImplementedError:
            self.log('Vector has no __eq__')
            return

        with fail_counter(
            test_name='Verify behavior of `element1 == element2`',
            logger=self.log
        ) as counter:

            for [n_x, x], [n_y, y] in samples(self.space,
                                              self.space):
                if n_x == n_y:
                    if not x == y:
                        counter.fail('failed x == x with x={:25s}'
                                     ''.format(n_x))

                    if x != y:
                        counter.fail('failed not x != x with x={:25s}'
                                     ''.format(n_x))
                else:
                    if x == y:
                        counter.fail('failed not x == y with x={:25s}, '
                                     'x={:25s}'.format(n_x, n_y))

                    if not x != y:
                        counter.fail('failed x != y with x={:25s}, x={:25s}'
                                     ''.format(n_x, n_y))

    def element_space(self):
        """Verify `LinearSpaceElement.space`."""
        with fail_counter(
            test_name='Verify `LinearSpaceElement.space`',
            logger=self.log
        ) as counter:

            for [n_x, x] in samples(self.space):
                if x.space != self.space:
                    counter.fail('failed with x={:25s}'.format(n_x))

    def element(self):
        """Verify `LinearSpaceElement`."""

        self.log('\n== Verifying element attributes ==\n')
        self.element_assign()
        self.element_copy()
        self.element_set_zero()
        self.element_equals()
        self.element_space()

    def run_tests(self):
        """Run all tests on this space."""
        self.log('\n== RUNNING ALL TESTS ==\n')
        self.log('Space = {}'.format(self.space))
        self.field()
        self.element_method()
        self.linearity()
        self.inner()
        self.norm()
        self.dist()
        self.multiply()
        self.equals()
        self.contains()
        self.element()

    def __str__(self):
        """Return ``str(self)``."""
        return '{}({})'.format(self.__class__.__name__, self.space)

    def __repr__(self):
        """Return ``repr(self)``."""
        inner_str = '{!r}'.format(self.space)
        if not self.verbose:
            inner_str += ', verbose=False'
        if self.tol != 1e-5:
            inner_str += ', tol={}'.format(self.tol)
        return '{}({})'.format(self.__class__.__name__, inner_str)


if __name__ == '__main__':
    from odl import rn, uniform_discr
    SpaceTest(rn(10), verbose=False).run_tests()
    SpaceTest(uniform_discr([0, 0], [1, 1], [5, 5])).run_tests()



# Copyright 2014-2017 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import division
from itertools import product
import numpy as np
import pytest

import odl
from odl.trafos.util.ft_utils import (
    reciprocal_grid, realspace_grid, dft_preprocess_data)
from odl.core.util import all_almost_equal, all_equal
from odl.core.util.testutils import simple_fixture


# --- pytest fixtures --- #


halfcomplex = simple_fixture('halfcomplex', [True, False])
shift = simple_fixture('shift', [True, False])
parity = simple_fixture('parity', ['even', 'odd'])
sign = simple_fixture('sign', ['-', '+'])


# --- reciprocal_grid --- #


def test_reciprocal_grid_1d(halfcomplex, shift, parity):

    shape = 10 if parity == 'even' else 11
    grid = odl.uniform_grid(0, 1, shape=shape)
    s = grid.stride
    n = np.array(grid.shape)

    rgrid = reciprocal_grid(grid, shift=shift, halfcomplex=halfcomplex)

    # Independent of halfcomplex, shift and parity
    true_recip_stride = 2 * np.pi / (s * n)
    assert all_almost_equal(rgrid.stride, true_recip_stride)

    if halfcomplex:
        assert all_equal(rgrid.shape, n // 2 + 1)

        if parity == 'odd' and shift:
            # Max point should be half a negative recip stride
            assert all_almost_equal(rgrid.max_pt, -true_recip_stride / 2)
        elif parity == 'even' and not shift:
            # Max point should be half a positive recip stride
            assert all_almost_equal(rgrid.max_pt, true_recip_stride / 2)
        elif (parity == 'odd' and not shift) or (parity == 'even' and shift):
            # Max should be zero
            assert all_almost_equal(rgrid.max_pt, 0)
        else:
            raise RuntimeError('parameter combination not covered')
    else:  # halfcomplex = False
        assert all_equal(rgrid.shape, n)

        if (parity == 'even' and shift) or (parity == 'odd' and not shift):
            # Zero should be at index n // 2
            assert all_almost_equal(rgrid[n // 2], 0)
        elif (parity == 'odd' and shift) or (parity == 'even' and not shift):
            # No point should be closer to 0 than half a recip stride
            atol = 0.999 * true_recip_stride / 2
            assert not rgrid.approx_contains(0, atol=atol)
        else:
            raise RuntimeError('parameter combination not covered')

        if not shift:
            # Grid Should be symmetric
            assert all_almost_equal(rgrid.min_pt, -rgrid.max_pt)
            if parity == 'odd':
                # Midpoint should be 0
                assert all_almost_equal(rgrid.mid_pt, 0)

    # Inverting should give back the original
    irgrid = realspace_grid(rgrid, grid.min_pt, halfcomplex=halfcomplex,
                            halfcx_parity=parity)
    assert irgrid.approx_equals(grid, atol=1e-6)


def test_reciprocal_grid_nd():

    grid = odl.uniform_grid([0] * 3, [1] * 3, shape=(3, 4, 5))
    s = grid.stride
    n = np.array(grid.shape)

    true_recip_stride = 2 * np.pi / (s * n)

    # Without shift altogether
    rgrid = reciprocal_grid(grid, shift=False, halfcomplex=False)

    assert all_equal(rgrid.shape, n)
    assert all_almost_equal(rgrid.stride, true_recip_stride)
    assert all_almost_equal(rgrid.min_pt, -rgrid.max_pt)

    # Inverting should give back the original
    irgrid = realspace_grid(rgrid, grid.min_pt, halfcomplex=False)
    assert irgrid.approx_equals(grid, atol=1e-6)


def test_reciprocal_grid_nd_shift_list():

    grid = odl.uniform_grid([0] * 3, [1] * 3, shape=(3, 4, 5))
    s = grid.stride
    n = np.array(grid.shape)
    shift = [False, True, False]

    true_recip_stride = 2 * np.pi / (s * n)

    # Shift only the even dimension, then zero must be contained
    rgrid = reciprocal_grid(grid, shift=shift, halfcomplex=False)
    noshift = np.where(np.logical_not(shift))

    assert all_equal(rgrid.shape, n)
    assert all_almost_equal(rgrid.stride, true_recip_stride)
    assert all_almost_equal(rgrid.min_pt[noshift], -rgrid.max_pt[noshift])
    assert all_almost_equal(rgrid[n // 2], [0] * 3)

    # Inverting should give back the original
    irgrid = realspace_grid(rgrid, grid.min_pt, halfcomplex=False)
    assert irgrid.approx_equals(grid, atol=1e-6)


def test_reciprocal_grid_nd_axes():

    grid = odl.uniform_grid([0] * 3, [1] * 3, shape=(3, 4, 5))
    s = grid.stride
    n = np.array(grid.shape)
    axes_list = [[1, -1], [0], 0, [0, 2, 1], [2, 0]]

    for axes in axes_list:
        active = np.zeros(grid.ndim, dtype=bool)
        active[axes] = True
        inactive = np.logical_not(active)

        true_recip_stride = np.empty(grid.ndim)
        true_recip_stride[active] = 2 * np.pi / (s[active] * n[active])
        true_recip_stride[inactive] = s[inactive]

        # Without shift altogether
        rgrid = reciprocal_grid(grid, shift=False, axes=axes,
                                halfcomplex=False)

        assert all_equal(rgrid.shape, n)
        assert all_almost_equal(rgrid.stride, true_recip_stride)
        assert all_almost_equal(rgrid.min_pt[active], -rgrid.max_pt[active])
        assert all_equal(rgrid.min_pt[inactive], grid.min_pt[inactive])
        assert all_equal(rgrid.max_pt[inactive], grid.max_pt[inactive])

        # Inverting should give back the original
        irgrid = realspace_grid(rgrid, grid.min_pt, axes=axes,
                                halfcomplex=False)
        assert irgrid.approx_equals(grid, atol=1e-6)


def test_reciprocal_grid_nd_halfcomplex():

    grid = odl.uniform_grid([0] * 3, [1] * 3, shape=(3, 4, 5))
    s = grid.stride
    n = np.array(grid.shape)
    stride_last = 2 * np.pi / (s[-1] * n[-1])
    n[-1] = n[-1] // 2 + 1

    # Without shift
    rgrid = reciprocal_grid(grid, shift=False, halfcomplex=True)
    assert all_equal(rgrid.shape, n)
    assert rgrid.max_pt[-1] == 0  # last dim is odd

    # With shift
    rgrid = reciprocal_grid(grid, shift=True, halfcomplex=True)
    assert all_equal(rgrid.shape, n)
    assert rgrid.max_pt[-1] == -stride_last / 2

    # Inverting should give back the original
    irgrid = realspace_grid(rgrid, grid.min_pt, halfcomplex=True,
                            halfcx_parity='odd')
    assert irgrid.approx_equals(grid, atol=1e-6)

    with pytest.raises(ValueError):
        realspace_grid(rgrid, grid.min_pt, halfcomplex=True,
                       halfcx_parity='+')


# --- dft_preprocess_data --- #


def test_dft_preprocess_data(sign):

    shape = (2, 3, 4)

    # With shift
    correct_arr = []
    for i, j, k in product(range(shape[0]), range(shape[1]), range(shape[2])):
        correct_arr.append((1 + 1j) * (1 - 2 * ((i + j + k) % 2)))

    arr = np.ones(shape, dtype='complex64') * (1 + 1j)
    preproc = dft_preprocess_data(arr, shift=True, sign=sign)  # out-of-place
    dft_preprocess_data(arr, shift=True, out=arr, sign=sign)  # in-place

    assert all_almost_equal(preproc.ravel(), correct_arr)
    assert all_almost_equal(arr.ravel(), correct_arr)

    # Without shift
    imag = 1j if sign == '-' else -1j
    correct_arr = []
    for i, j, k in product(range(shape[0]), range(shape[1]), range(shape[2])):
        argsum = sum((idx * (1 - 1 / shp))
                     for idx, shp in zip((i, j, k), shape))

        correct_arr.append((1 + 1j) * np.exp(imag * np.pi * argsum))

    arr = np.ones(shape, dtype='complex64') * (1 + 1j)
    dft_preprocess_data(arr, shift=False, out=arr, sign=sign)

    assert all_almost_equal(arr.ravel(), correct_arr)

    # Bad input
    with pytest.raises(ValueError):
        dft_preprocess_data(arr, out=arr, sign=1)

    arr = np.zeros(shape, dtype='S2')
    with pytest.raises(ValueError):
        dft_preprocess_data(arr)


def test_dft_preprocess_data_halfcomplex(sign):

    shape = (2, 3, 4)

    # With shift
    correct_arr = []
    for i, j, k in product(range(shape[0]), range(shape[1]), range(shape[2])):
        correct_arr.append(1 - 2 * ((i + j + k) % 2))

    arr = np.ones(shape, dtype='float64')
    preproc = dft_preprocess_data(arr, shift=True, sign=sign)  # out-of-place
    out = np.empty_like(arr)
    dft_preprocess_data(arr, shift=True, out=out, sign=sign)  # in-place
    dft_preprocess_data(arr, shift=True, out=arr, sign=sign)  # in-place
    assert all_almost_equal(preproc.ravel(), correct_arr)
    assert all_almost_equal(arr.ravel(), correct_arr)
    assert all_almost_equal(out.ravel(), correct_arr)

    # Without shift
    imag = 1j if sign == '-' else -1j
    correct_arr = []
    for i, j, k in product(range(shape[0]), range(shape[1]), range(shape[2])):
        argsum = sum((idx * (1 - 1 / shp))
                     for idx, shp in zip((i, j, k), shape))

        correct_arr.append(np.exp(imag * np.pi * argsum))

    arr = np.ones(shape, dtype='float64')
    preproc = dft_preprocess_data(arr, shift=False, sign=sign)
    assert all_almost_equal(preproc.ravel(), correct_arr)

    # Non-float input works, too
    arr = np.ones(shape, dtype='int')
    preproc = dft_preprocess_data(arr, shift=False, sign=sign)
    assert all_almost_equal(preproc.ravel(), correct_arr)

    # In-place modification not possible for float array and no shift
    arr = np.ones(shape, dtype='float64')
    with pytest.raises(ValueError):
        dft_preprocess_data(arr, shift=False, out=arr, sign=sign)


def test_dft_preprocess_data_with_axes(sign):

    shape = (2, 3, 4)

    axes = 1  # Only middle index counts
    correct_arr = []
    for _, j, __ in product(range(shape[0]), range(shape[1]), range(shape[2])):
        correct_arr.append(1 - 2 * (j % 2))

    arr = np.ones(shape, dtype='complex64')
    dft_preprocess_data(arr, shift=True, axes=axes, out=arr, sign=sign)

    assert all_almost_equal(arr.ravel(), correct_arr)

    axes = [0, -1]  # First and last
    correct_arr = []
    for i, _, k in product(range(shape[0]), range(shape[1]), range(shape[2])):
        correct_arr.append(1 - 2 * ((i + k) % 2))

    arr = np.ones(shape, dtype='complex64')
    dft_preprocess_data(arr, shift=True, axes=axes, out=arr, sign=sign)

    assert all_almost_equal(arr.ravel(), correct_arr)


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2025 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Python Array API support."""

from __future__ import absolute_import

from .element_wise import *
from .statistical import *
from .utils import *
from .comparisons import *
from .array_creation import *

__all__ = ()
__all__ += element_wise.__all__
__all__ += statistical.__all__
__all__ += utils.__all__
__all__ += comparisons.__all__
__all__ += array_creation.__all__



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Test analytical reconstruction methods."""

from __future__ import division

import numpy as np
import pytest

import odl
import odl.applications.tomo as tomo
from odl.applications.tomo.util.testutils import (
    skip_if_no_astra, skip_if_no_astra_cuda, skip_if_no_skimage)
from odl.core.util.testutils import simple_fixture

# --- pytest fixtures --- #


# Global pytest mark
pytestmark = pytest.mark.suite('largescale')


filter_type = simple_fixture(
    'filter_type', ['Ram-Lak', 'Shepp-Logan', 'Cosine', 'Hamming', 'Hann'])
frequency_scaling = simple_fixture(
    'frequency_scaling', [0.5, 0.9, 1.0])

weighting = simple_fixture('weighting', [None, 1.0])

# Find the valid projectors
# TODO: Add nonuniform once #671 is solved
projectors = []
projectors.extend(
    (pytest.param(proj_cfg, marks=skip_if_no_astra)
     for proj_cfg in ['par2d astra_cpu uniform',
                      'cone2d astra_cpu uniform']
     )
)
projectors.extend(
    (pytest.param(proj_cfg, marks=skip_if_no_astra_cuda)
     for proj_cfg in ['par2d astra_cuda uniform',
                      'cone2d astra_cpu uniform',
                      'cone2d astra_cuda uniform',
                      'par3d astra_cuda uniform',
                      'cone3d astra_cuda uniform',
                      'helical astra_cuda uniform']
     )
)
projectors.extend(
    (pytest.param(proj_cfg, marks=skip_if_no_skimage)
     for proj_cfg in ['par2d skimage uniform']
     )
)
projector_ids = [
    " geom='{}' - impl='{}' - angles='{}' ".format(*p.values[0].split())
    for p in projectors
]


@pytest.fixture(scope="module", params=projectors, ids=projector_ids)
def projector(request, weighting):

    n_angles = 500
    dtype = 'float32'

    geom, impl, angle = request.param.split()

    if angle == 'uniform':
        apart = odl.uniform_partition(0, 2 * np.pi, n_angles)
    elif angle == 'random':
        # Linearly spaced with random noise
        min_pt = 2 * (2.0 * np.pi) / n_angles
        max_pt = (2.0 * np.pi) - 2 * (2.0 * np.pi) / n_angles
        points = np.linspace(min_pt, max_pt, n_angles)
        points += np.random.rand(n_angles) * (max_pt - min_pt) / (5 * n_angles)
        apart = odl.nonuniform_partition(points)
    elif angle == 'nonuniform':
        # Angles spaced quadratically
        min_pt = 2 * (2.0 * np.pi) / n_angles
        max_pt = (2.0 * np.pi) - 2 * (2.0 * np.pi) / n_angles
        points = np.linspace(min_pt ** 0.5, max_pt ** 0.5, n_angles) ** 2
        apart = odl.nonuniform_partition(points)
    else:
        raise ValueError('angle not valid')

    if geom == 'par2d':
        # Reconstruction space
        discr_reco_space = odl.uniform_discr([-20, -20], [20, 20],
                                             [100, 100], dtype=dtype,
                                             weighting=weighting)

        # Geometry
        dpart = odl.uniform_partition(-30, 30, 500)
        geom = tomo.Parallel2dGeometry(apart, dpart)

        # Ray transform
        return tomo.RayTransform(discr_reco_space, geom, impl=impl)

    elif geom == 'par3d':
        # Reconstruction space
        discr_reco_space = odl.uniform_discr([-20, -20, -20], [20, 20, 20],
                                             [100, 100, 100], dtype=dtype,
                                             weighting=weighting)

        # Geometry
        dpart = odl.uniform_partition([-30, -30], [30, 30], [200, 200])
        geom = tomo.Parallel3dAxisGeometry(apart, dpart, axis=[1, 1, 0])

        # Ray transform
        return tomo.RayTransform(discr_reco_space, geom, impl=impl)

    elif geom == 'cone2d':
        # Reconstruction space
        discr_reco_space = odl.uniform_discr([-20, -20], [20, 20],
                                             [100, 100], dtype=dtype,
                                             weighting=weighting)

        # Geometry
        dpart = odl.uniform_partition(-40, 40, 200)
        geom = tomo.FanBeamGeometry(apart, dpart,
                                    src_radius=100, det_radius=100)

        # Ray transform
        return tomo.RayTransform(discr_reco_space, geom, impl=impl)

    elif geom == 'cone3d':
        # Reconstruction space
        discr_reco_space = odl.uniform_discr([-20, -20, -20], [20, 20, 20],
                                             [100, 100, 100], dtype=dtype,
                                             weighting=weighting)

        # Geometry
        dpart = odl.uniform_partition([-50, -50], [50, 50], [200, 200])
        geom = tomo.ConeBeamGeometry(
            apart, dpart, src_radius=100, det_radius=100, axis=[1, 0, 0])

        # Ray transform
        return tomo.RayTransform(discr_reco_space, geom, impl=impl)

    elif geom == 'helical':
        # Reconstruction space
        discr_reco_space = odl.uniform_discr([-20, -20, 0], [20, 20, 40],
                                             [100, 100, 100], dtype=dtype,
                                             weighting=weighting)

        # Geometry
        # TODO: angles
        n_angle = 2000
        apart = odl.uniform_partition(0, 8 * 2 * np.pi, n_angle)
        dpart = odl.uniform_partition([-50, -4], [50, 4], [200, 20])
        geom = tomo.ConeBeamGeometry(
            apart, dpart, src_radius=100, det_radius=100, pitch=5.0)

        # Windowed ray transform
        return tomo.RayTransform(discr_reco_space, geom, impl=impl)
    else:
        raise ValueError('param not valid')


# --- FBP tests --- #


def test_fbp_reconstruction(projector):
    """Test filtered back-projection with various projectors."""

    # Create Shepp-Logan phantom
    vol = odl.core.phantom.shepp_logan(projector.domain, modified=False)

    # Project data
    projections = projector(vol)

    # Create default FBP operator and apply to projections
    fbp_operator = odl.applications.tomo.fbp_op(projector)

    # Add window if problem is in 3d.
    if (
        isinstance(projector.geometry, odl.applications.tomo.ConeBeamGeometry)
        and projector.geometry.pitch != 0
    ):
        fbp_operator = fbp_operator * odl.applications.tomo.tam_danielson_window(projector)

    # Compute the FBP result
    fbp_result = fbp_operator(projections)

    # Allow 30 % error
    maxerr = vol.norm() * 0.3
    error = vol.dist(fbp_result)
    assert error < maxerr


@skip_if_no_astra_cuda
def test_fbp_reconstruction_filters(filter_type, frequency_scaling, weighting):
    """Validate that the various filters work as expected."""

    apart = odl.uniform_partition(0, np.pi, 500)

    discr_reco_space = odl.uniform_discr([-20, -20], [20, 20],
                                         [100, 100], dtype='float32',
                                         weighting=weighting)

    # Geometry
    dpart = odl.uniform_partition(-30, 30, 500)
    geom = tomo.Parallel2dGeometry(apart, dpart)

    # Ray transform
    projector = tomo.RayTransform(discr_reco_space, geom, impl='astra_cuda')

    # Create Shepp-Logan phantom
    vol = odl.core.phantom.shepp_logan(projector.domain, modified=False)

    # Project data
    projections = projector(vol)

    # Create FBP operator with filters and apply to projections
    fbp_operator = odl.applications.tomo.fbp_op(projector,
                                   filter_type=filter_type,
                                   frequency_scaling=frequency_scaling)

    fbp_result = fbp_operator(projections)

    maxerr = vol.norm() / 5.0
    error = vol.dist(fbp_result)
    assert error < maxerr


if __name__ == '__main__':
    odl.core.util.test_file(__file__, ['-S', 'largescale'])



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""ODL integration with pyshearlab."""


from threading import Lock

import numpy as np

import odl
import pyshearlab_mind as pyshearlab

__all__ = ('PyShearlabOperator',)


class PyShearlabOperator(odl.Operator):

    """Shearlet transform using PyShearlab.

    This is the non-compact shearlet transform implemented using the fourier
    transform.
    """

    def __init__(self, space, num_scales):
        """Initialize a new instance.

        Parameters
        ----------
        space : `DiscretizedSpace`
            The space on which the shearlet transform should act. Must be
            two-dimensional.
        num_scales : nonnegative `int`
            The number of scales for the shearlet transform, higher numbers
            mean better edge resolution but more computational burden.

        Examples
        --------
        Create a 2d-shearlet transform:

        >>> space = odl.uniform_discr([-1, -1], [1, 1], [128, 128])
        >>> shearlet_transform = PyShearlabOperator(space, num_scales=2)
        """
        assert(space.impl=='numpy')
        self.shearlet_system = pyshearlab.SLgetShearletSystem2D(
            0, space.shape[0], space.shape[1], num_scales)
        range = space ** self.shearlet_system['nShearlets']
        self.mutex = Lock()
        super(PyShearlabOperator, self).__init__(space, range, True)

    def _call(self, x):
        """Return ``self(x)``."""
        with self.mutex:
            result = pyshearlab.SLsheardec2D(x.asarray(), self.shearlet_system)
            return self.range.element(np.moveaxis(result, -1, 0))

    @property
    def adjoint(self):
        """Return the adjoint operator."""
        return PyShearlabOperatorAdjoint(self)

    @property
    def inverse(self):
        """Return the inverse operator."""
        return PyShearlabOperatorInverse(self)


class PyShearlabOperatorAdjoint(odl.Operator):

    """Adjoint of the shearlet transform.

    Should not be used independently.

    See Also
    --------
    odl.contrib.pyshearlab.PyShearlabOperator
    """

    def __init__(self, op):
        """Initialize a new instance.

        Parameters
        ----------
        op : `PyShearlabOperator`
            The operator which this should be the adjoint of.
        """
        self.op = op
        super(PyShearlabOperatorAdjoint, self).__init__(
            op.range, op.domain, True)

    def _call(self, x):
        """Return ``self(x)``."""
        with self.op.mutex:
            x = np.moveaxis(x.asarray(), 0, -1)
            return self.range.element(pyshearlab.SLshearadjoint2D(x, self.op.shearlet_system))

    @property
    def adjoint(self):
        """Return the adjoint operator."""
        return self.op

    @property
    def inverse(self):
        """Return the inverse operator."""
        return PyShearlabOperatorAdjointInverse(self.op)


class PyShearlabOperatorInverse(odl.Operator):

    """Inverse of the shearlet transform.

    Should not be used independently.

    See Also
    --------
    odl.contrib.pyshearlab.PyShearlabOperator
    """

    def __init__(self, op):
        """Initialize a new instance.

        Parameters
        ----------
        op : `PyShearlabOperator`
            The operator which this should be the inverse of.
        """
        self.op = op
        super(PyShearlabOperatorInverse, self).__init__(
            op.range, op.domain, True)

    def _call(self, x):
        """Return ``self(x)``."""
        with self.op.mutex:
            x = np.moveaxis(x.asarray(), 0, -1)
            return self.range.element(pyshearlab.SLshearrec2D(x, self.op.shearlet_system))

    @property
    def adjoint(self):
        """Return the adjoint operator."""
        return PyShearlabOperatorAdjointInverse(self.op)

    @property
    def inverse(self):
        """Return the inverse operator."""
        return self.op


class PyShearlabOperatorAdjointInverse(odl.Operator):

    """Adjoint of the inverse/Inverse of the adjoint of shearlet transform.

    Should not be used independently.

    See Also
    --------
    odl.contrib.pyshearlab.PyShearlabOperator
    """

    def __init__(self, op):
        """Initialize a new instance.

        Parameters
        ----------
        op : `PyShearlabOperator`
            The operator which this should be the inverse of the adjoint of.
        """
        self.op = op
        super(PyShearlabOperatorAdjointInverse, self).__init__(
            op.domain, op.range, True)

    def _call(self, x):
        """Return ``self(x)``."""
        with self.op.mutex:
            result = pyshearlab.SLshearrecadjoint2D(x.asarray(), self.op.shearlet_system)
            return self.range.element(np.moveaxis(result, -1, 0))

    @property
    def adjoint(self):
        """Return the adjoint operator."""
        return self.op.inverse

    @property
    def inverse(self):
        """Return the inverse operator."""
        return self.op.adjoint


if __name__ == '__main__':
    # pylint: disable=wrong-import-position
    from odl.core.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Operators and functions for linearized deformation."""

from __future__ import absolute_import, division, print_function

import numpy as np

from odl.core.discr import DiscretizedSpace, Divergence, Gradient
from odl.core.discr.discr_space import DiscretizedSpaceElement
from odl.core.discr.discr_utils import _normalize_interp, per_axis_interpolator
from odl.core.operator import Operator, PointwiseInner
from odl.core.space import ProductSpace
from odl.core.space.pspace import ProductSpaceElement
from odl.core.util import indent, signature_string
from odl.core.array_API_support import exp, lookup_array_backend

__all__ = ('LinDeformFixedTempl', 'LinDeformFixedDisp', 'linear_deform')


def linear_deform(template, displacement, interp='linear', out=None):
    """Linearized deformation of a template with a displacement field.

    The function maps a given template ``I`` and a given displacement
    field ``v`` to the new function ``x --> I(x + v(x))``.

    Parameters
    ----------
    template : `DiscretizedSpaceElement`
        Template to be deformed by a displacement field.
    displacement : element of power space of ``template.space``
        Vector field (displacement field) used to deform the
        template.
    interp : str or sequence of str
        Interpolation type that should be used to sample the template on
        the deformed grid. A single value applies to all axes, and a
        sequence gives the interpolation scheme per axis.

        Supported values: ``'nearest'``, ``'linear'``

    out : `numpy.ndarray`, optional
        Array to which the function values of the deformed template
        are written. It must have the same shape as ``template`` and
        a data type compatible with ``template.dtype``.

    Returns
    -------
    deformed_template : `numpy.ndarray`
        Function values of the deformed template. If ``out`` was given,
        the returned object is a reference to it.

    Examples
    --------
    Create a simple 1D template to initialize the operator and
    apply it to a displacement field. Where the displacement is zero,
    the output value is the same as the input value.
    In the 4-th point, the value is taken from 0.2 (one cell) to the
    left, i.e. 1.0.

    >>> space = odl.uniform_discr(0, 1, 5)
    >>> disp_field_space = space.tangent_bundle
    >>> template = space.element([0, 0, 1, 0, 0])
    >>> displacement_field = disp_field_space.element([[0, 0, 0, -0.2, 0]])
    >>> linear_deform(template, displacement_field, interp='nearest')
    array([ 0.,  0.,  1.,  1.,  0.])

    The result depends on the chosen interpolation. With 'linear'
    interpolation and an offset of half the distance between two
    points, 0.1, one gets the mean of the values.

    >>> displacement_field = disp_field_space.element([[0, 0, 0, -0.1, 0]])
    >>> linear_deform(template, displacement_field, interp='linear')
    array([ 0. ,  0. ,  1. ,  0.5,  0. ])
    """
    points = template.space.points()
    if isinstance(displacement, ProductSpaceElement):
        impl, device = displacement[0].impl, displacement[0].device
        backend = lookup_array_backend(impl)
    else:
        raise ValueError(f'{type(displacement)}')
    
    points = backend.array_constructor(points, device=device)
    
    for i, vi in enumerate(displacement):
        points[:, i] += vi.asarray().ravel()
    templ_interpolator = per_axis_interpolator(
        template, coord_vecs=template.space.grid.coord_vectors, interp=interp
    )

    values = templ_interpolator(points.T, out=out)
    return values.reshape(template.space.shape)


class LinDeformFixedTempl(Operator):

    r"""Deformation operator with fixed template acting on displacement fields.

    The operator has a fixed template ``I`` and maps a displacement
    field ``v`` to the new function ``x --> I(x + v(x))``.

    See Also
    --------
    LinDeformFixedDisp : Deformation with a fixed displacement.

    Notes
    -----
    For :math:`\Omega \subset \mathbb{R}^d`, we take :math:`X = L^p(\Omega)`
    to be the template space, i.e. :math:`I \in X`. Then the vector field
    space is identified with :math:`V := X^d`. Hence the deformation operator
    with fixed template maps :math:`V` into :math:`X`:

    .. math::
        W_I : V \to X, \quad W_I(v) := I(\cdot + v(\cdot)),

    i.e., :math:`W_I(v)(x) = I(x + v(x))`.

    Note that this operator is non-linear. Its derivative at :math:`v` is
    an operator that maps :math:`V` into :math:`X`:

    .. math::
        W_I'(v) : V \to X, \quad W_I'(v)(u) =
        \big< \nabla I(\cdot + v(\cdot)), u \big>_{\mathbb{R}^d},

    i.e., :math:`W_I'(v)(u)(x) = \nabla I(x + v(x))^T u(x)`,

    which is to be understood as a point-wise inner product, resulting
    in a function in :math:`X`. And the adjoint of the preceding derivative
    is also an operator that maps :math:`X` into :math:`V`:

    .. math::
        W_I'(v)^* : X \to V, \quad W_I'(v)^*(J) =
        J \, \nabla I(\cdot + v(\cdot)),

    i.e., :math:`W_I'(v)^*(J)(x) = J(x) \, \nabla I(x + v(x))`.
    """

    def __init__(self, template, domain=None, interp='linear'):
        """Initialize a new instance.

        Parameters
        ----------
        template : `DiscretizedSpaceElement`
            Fixed template that is to be deformed.
        domain : power space of `DiscretizedSpace`, optional
            The space of all allowed coordinates in the deformation.
            A `ProductSpace` of ``template.ndim`` copies of a function-space.
            It must fulfill
            ``domain[0].partition == template.space.partition``, so
            this option is useful mainly when using different interpolations
            in displacement and template.

            Default: ``template.space.real_space.tangent_bundle``

        interp : str or sequence of str
            Interpolation type that should be used to sample the template on
            the deformed grid. A single value applies to all axes, and a
            sequence gives the interpolation scheme per axis.

            Supported values: ``'nearest'``, ``'linear'``

            .. warning::
                Choosing ``'nearest'`` interpolation results in a formally
                non-differentiable operator since the gradient of the
                template is not well-defined. If the operator derivative
                is to be used, a differentiable interpolation scheme (e.g.,
                ``'linear'``) should be chosen.

        Examples
        --------
        Create a simple 1D template to initialize the operator and
        apply it to a displacement field. Where the displacement is zero,
        the output value is the same as the input value.
        In the 4-th point, the value is taken from 0.2 (one cell) to the
        left, i.e. 1.0.

        >>> space = odl.uniform_discr(0, 1, 5)
        >>> template = space.element([0, 0, 1, 0, 0])
        >>> op = LinDeformFixedTempl(template, interp='nearest')
        >>> disp_field = [[0, 0, 0, -0.2, 0]]
        >>> print(op(disp_field))
        [ 0.,  0.,  1.,  1.,  0.]

        The result depends on the chosen interpolation. With 'linear'
        interpolation and an offset of half the distance between two
        points, 0.1, one gets the mean of the values.

        >>> op = LinDeformFixedTempl(template, interp='linear')
        >>> disp_field = [[0, 0, 0, -0.1, 0]]
        >>> print(op(disp_field))
        [ 0. ,  0. ,  1. ,  0.5,  0. ]
        """
        if not isinstance(template, DiscretizedSpaceElement):
            raise TypeError(
                '`template` must be a `DiscretizedSpaceElement, got {!r}`'
                ''.format(template)
            )
        self.__template = template

        if domain is None:
            domain = self.template.space.real_space.tangent_bundle
        else:
            if not isinstance(domain, ProductSpace):
                # TODO: allow non-product spaces in the 1D case
                raise TypeError('`domain` must be a `ProductSpace` '
                                'instance, got {!r}'.format(domain))
            if not domain.is_power_space:
                raise TypeError('`domain` must be a power space, '
                                'got {!r}'.format(domain))
            if not isinstance(domain[0], DiscretizedSpace):
                raise TypeError('`domain[0]` must be a `DiscretizedSpace` '
                                'instance, got {!r}'.format(domain[0]))

            if template.space.partition != domain[0].partition:
                raise ValueError(
                    '`template.space.partition` not equal to `coord_space`s '
                    'partiton ({!r} != {!r})'
                    ''.format(template.space.partition, domain[0].partition))

        super(LinDeformFixedTempl, self).__init__(
            domain=domain, range=template.space, linear=False)

        self.__interp_byaxis = _normalize_interp(interp, template.space.ndim)

    @property
    def template(self):
        """Fixed template of this deformation operator."""
        return self.__template

    @property
    def interp_byaxis(self):
        """Tuple of per-axis interpolation schemes."""
        return self.__interp_byaxis

    @property
    def interp(self):
        """Interpolation scheme or tuple of per-axis interpolation schemes."""
        if (
            len(self.interp_byaxis) != 0
            and all(s == self.interp_byaxis[0] for s in self.interp_byaxis[1:])
        ):
            return self.interp_byaxis[0]
        else:
            return self.interp_byaxis

    def _call(self, displacement, out=None):
        """Implementation of ``self(displacement[, out])``."""
        return linear_deform(self.template, displacement, self.interp, out)

    def derivative(self, displacement):
        """Derivative of the operator at ``displacement``.

        Parameters
        ----------
        displacement : `domain` `element-like`
            Point at which the derivative is computed.

        Returns
        -------
        derivative : `PointwiseInner`
            The derivative evaluated at ``displacement``.
        """
        # To implement the complex case we need to be able to embed the real
        # vector field space into the range of the gradient. Issue #59.
        if not self.range.is_real:
            raise NotImplementedError('derivative not implemented for complex '
                                      'spaces.')

        displacement = self.domain.element(displacement)

        # TODO: allow users to select what method to use here.
        grad = Gradient(domain=self.range, method='central',
                        pad_mode='symmetric')
        grad_templ = grad(self.template)
        def_grad = self.domain.element(
            [linear_deform(gf, displacement, self.interp) for gf in grad_templ]
        )

        return PointwiseInner(self.domain, def_grad)

    def __repr__(self):
        """Return ``repr(self)``."""
        posargs = [self.template]
        optargs = [
            ('domain', self.domain, self.template.space.tangent_bundle),
            ('interp', self.interp, 'linear'),
        ]
        inner_str = signature_string(posargs, optargs, mod='!r', sep=',\n')
        return '{}(\n{}\n)'.format(self.__class__.__name__, indent(inner_str))


class LinDeformFixedDisp(Operator):

    r"""Deformation operator with fixed displacement acting on templates.

    The operator has a fixed displacement field ``v`` and maps a template
    ``I`` to the new function ``x --> I(x + v(x))``.

    See Also
    --------
    LinDeformFixedTempl : Deformation with a fixed template.

    Notes
    -----
    For :math:`\Omega \subset \mathbb{R}^d`, we take :math:`V := X^d`
    to be the space of displacement fields, where :math:`X = L^p(\Omega)`
    is the template space. Hence the deformation operator with the fixed
    displacement field :math:`v \in V` maps :math:`X` into :math:`X`:

    .. math::
        W_v : X \to X, \quad W_v(I) := I(\cdot + v(\cdot)),

    i.e., :math:`W_v(I)(x) = I(x + v(x))`.

    This operator is linear, so its derivative is itself, but it may not be
    bounded and may thus not have a formal adjoint. For "small" :math:`v`,
    though, one can approximate the adjoint by

    .. math::
        W_v^*(I) \approx \exp(-\mathrm{div}\, v) \, I(\cdot - v(\cdot)),

    i.e., :math:`W_v^*(I)(x) \approx \exp(-\mathrm{div}\,v(x))\, I(x - v(x))`.
    """

    def __init__(self, displacement, templ_space=None, interp='linear'):
        """Initialize a new instance.

        Parameters
        ----------
        displacement : element of a power space of `DiscretizedSpace`
            Fixed displacement field used in the deformation.
        templ_space : `DiscretizedSpace`, optional
            Template space on which this operator is applied, i.e. the
            operator domain and range. It must fulfill
            ``templ_space[0].partition == displacement.space.partition``, so
            this option is useful mainly for support of complex spaces and if
            different interpolations should be used for displacement and
            template.

            Default: ``displacement.space[0]``

        interp : str or sequence of str
            Interpolation type that should be used to sample the template on
            the deformed grid. A single value applies to all axes, and a
            sequence gives the interpolation scheme per axis.

            Supported values: ``'nearest'``, ``'linear'``

        Examples
        --------
        Create a simple 1D template to initialize the operator and
        apply it to a displacement field. Where the displacement is zero,
        the output value is the same as the input value.
        In the 4-th point, the value is taken from 0.2 (one cell) to the
        left, i.e. 1.0.

        >>> space = odl.uniform_discr(0, 1, 5)
        >>> disp_field = space.tangent_bundle.element([[0, 0, 0, -0.2, 0]])
        >>> op = odl.trafos.deform.LinDeformFixedDisp(disp_field, interp='nearest')
        >>> template = [0, 0, 1, 0, 0]
        >>> print(op([0, 0, 1, 0, 0]))
        [ 0.,  0.,  1.,  1.,  0.]

        The result depends on the chosen interpolation. With 'linear'
        interpolation and an offset of half the distance between two
        points, 0.1, one gets the mean of the values.

        >>> disp_field = space.tangent_bundle.element([[0, 0, 0, -0.1, 0]])
        >>> op = odl.trafos.deform.LinDeformFixedDisp(disp_field, interp='linear')
        >>> template = [0, 0, 1, 0, 0]
        >>> print(op(template))
        [ 0. ,  0. ,  1. ,  0.5,  0. ]
        """
        if not isinstance(displacement, ProductSpaceElement):
            raise TypeError(
                '`displacement` must be a `ProductSpaceElement`, got {!r}'
                ''.format(displacement)
            )

        if not displacement.space.is_power_space:
            raise ValueError(
                '`displacement.space` must be a power space, got {!r}'
                ''.format(displacement.space)
            )
        if not isinstance(displacement.space[0], DiscretizedSpace):
            raise ValueError(
                '`displacement.space[0]` must be a `DiscretizedSpace`, '
                'got {!r}'.format(displacement.space[0]))

        self.__displacement = displacement

        if templ_space is None:
            templ_space = displacement.space[0]
        else:
            if not isinstance(templ_space, DiscretizedSpace):
                raise TypeError('`templ_space` must be a `DiscretizedSpace` '
                                'instance, got {!r}'.format(templ_space))
            if templ_space.partition != displacement.space[0].partition:
                raise ValueError(
                    '`templ_space.partition` not equal to `displacement`s '
                    'partiton ({!r} != {!r})'
                    ''.format(templ_space.partition,
                              displacement.space[0].partition)
                )

        super(LinDeformFixedDisp, self).__init__(
            domain=templ_space, range=templ_space, linear=True)

        self.__interp_byaxis = _normalize_interp(interp, templ_space.ndim)

    @property
    def interp_byaxis(self):
        """Tuple of per-axis interpolation schemes."""
        return self.__interp_byaxis

    @property
    def interp(self):
        """Interpolation scheme or tuple of per-axis interpolation schemes."""
        if (
            len(self.interp_byaxis) != 0
            and all(s == self.interp_byaxis[0] for s in self.interp_byaxis[1:])
        ):
            return self.interp_byaxis[0]
        else:
            return self.interp_byaxis

    @property
    def displacement(self):
        """Fixed displacement field of this deformation operator."""
        return self.__displacement

    def _call(self, template, out=None):
        """Implementation of ``self(template[, out])``."""
        return linear_deform(template, self.displacement, self.interp, out)

    @property
    def inverse(self):
        """Inverse deformation using ``-v`` as displacement.

        Note that this implementation uses an approximation that is only
        valid for small displacements.
        """
        return LinDeformFixedDisp(
            -self.displacement, templ_space=self.domain, interp=self.interp
        )

    @property
    def adjoint(self):
        """Adjoint of the linear operator.

        Note that this implementation uses an approximation that is only
        valid for small displacements.
        """
        # TODO allow users to select what method to use here.
        div_op = Divergence(domain=self.displacement.space, method='forward',
                            pad_mode='symmetric')
        jacobian_det = self.domain.element(exp(-div_op(self.displacement)))

        return jacobian_det * self.inverse

    def __repr__(self):
        """Return ``repr(self)``."""
        posargs = [self.displacement]
        optargs = [
            ('templ_space', self.domain, self.displacement.space[0]),
            ('interp', self.interp, 'linear'),
        ]
        inner_str = signature_string(posargs, optargs, mod='!r', sep=',\n')
        return '{}(\n{}\n)'.format(self.__class__.__name__, indent(inner_str))


if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Tests for linearized deformation operators."""

from __future__ import division

import numpy as np
import pytest

import odl
from odl.trafos.deform import LinDeformFixedDisp, LinDeformFixedTempl
from odl.core.util.testutils import simple_fixture

from odl.core.array_API_support import get_array_and_backend, exp

# --- pytest fixtures --- #


dtype = simple_fixture('dtype', [float, complex])
interp = simple_fixture('interp', ['linear', 'nearest'])
ndim = simple_fixture('ndim', [1, 2, 3])


@pytest.fixture
def space(ndim, dtype, odl_impl_device_pairs):
    """Provide a space for unit tests."""
    impl, device = odl_impl_device_pairs
    # supported_dtypes = odl.lookup_array_backend(impl).available_dtypes
    # # if np.dtype(dtype) not in supported_dtypes:
    # #     pytest.skip('dtype not available for this backend')

    return odl.uniform_discr(
        [-1] * ndim, [1] * ndim, [20] * ndim, impl=impl, dtype=dtype, device=device
    )


# --- Helper functions --- #


SIGMA = 0.3  # width of the gaussian
EPS = 0.25   # scale of the displacement field


def error_bound(interp):
    """Error bound varies with interpolation (larger for "worse")."""
    if interp == 'linear':
        return 0.1
    elif interp == 'nearest':
        return 0.2


def prod(x):
    """Product of a sequence."""
    prod = 1
    for xi in x:
        prod = prod * xi
    return prod


def numpy_template_function(x):
    """Gaussian function with std SIGMA."""
    return np.exp(-sum(xi ** 2 for xi in x) / SIGMA ** 2)

def torch_template_function(x):
    """Gaussian function with std SIGMA."""
    import torch
    return torch.exp(-sum(xi ** 2 for xi in x) / SIGMA ** 2)

def numpy_template_grad_factory(n):
    """Gradient of the gaussian."""
    def template_grad_i(i):
        # Indirection for lambda capture
        return lambda x: -2 * x[i] / SIGMA ** 2 * numpy_template_function(x)
    return [template_grad_i(i) for i in range(n)]

def torch_template_grad_factory(n):
    """Gradient of the gaussian."""
    def template_grad_i(i):
        # Indirection for lambda capture
        return lambda x: -2 * x[i] / SIGMA ** 2 * torch_template_function(x)
    return [template_grad_i(i) for i in range(n)]


def disp_field_factory(n):
    """Displacement field.

    In 1d: (x)
    In 2d: (xy, y)
    In 3d: (xyz, y, z)
    etc...
    """
    def coordinate_projection_i(i):
        # Indirection for lambda capture
        return lambda x: EPS * x[i]

    lst = [lambda x: EPS * prod(x)]
    lst.extend(coordinate_projection_i(i) for i in range(1, n))
    return lst


def numpy_exp_div_inv_disp(x):
    """Exponential of the divergence of the displacement field.

    In 1d: exp(- EPS)
    In 2d: exp(- EPS * (y + 1))
    In 2d: exp(- EPS * (yz + 2))
    """
    return np.exp(- EPS * (prod(x[1:]) + (len(x) - 1)))


def torch_exp_div_inv_disp(x):
    """Exponential of the divergence of the displacement field.

    In 1d: exp(- EPS)
    In 2d: exp(- EPS * (y + 1))
    In 2d: exp(- EPS * (yz + 2))
    """
    import torch
    return torch.exp(- EPS * (prod(x[1:]) + (len(x) - 1)))


def displaced_points(x):
    """Displaced coordinate points."""
    disp = [dsp(x) for dsp in disp_field_factory(len(x))]
    return [xi + di for xi, di in zip(x, disp)]


def numpy_deformed_template(x):
    """Deformed template."""
    return numpy_template_function(displaced_points(x))

def torch_deformed_template(x):
    """Deformed template."""
    return torch_template_function(displaced_points(x))

def vector_field_factory(n):
    """Vector field for the gradient.

    In 1d: (x)
    In 2d: (x, y)
    In 3d: (x, y, z)
    etc...
    """
    def vector_field_i(i):
        return lambda x: x[i]
    return [vector_field_i(i) for i in range(n)]


def numpy_template_deformed_grad_factory(n):
    """Deformed gradient."""
    templ_grad = numpy_template_grad_factory(n)

    def template_deformed_gradi(i):
        # Indirection for lambda capture
        return lambda x: templ_grad[i](displaced_points(x))

    return [template_deformed_gradi(i) for i in range(n)]

def torch_template_deformed_grad_factory(n):
    """Deformed gradient."""
    templ_grad = torch_template_grad_factory(n)

    def template_deformed_gradi(i):
        # Indirection for lambda capture
        return lambda x: templ_grad[i](displaced_points(x))

    return [template_deformed_gradi(i) for i in range(n)]

def numpy_fixed_templ_deriv(x):
    """Derivative taken in disp_field and evaluated in vector_field."""
    dg = [tdgf(x) for tdgf in numpy_template_deformed_grad_factory(len(x))]
    v = [vff(x) for vff in vector_field_factory(len(x))]
    return sum(dgi * vi for dgi, vi in zip(dg, v))

def torch_fixed_templ_deriv(x):
    """Derivative taken in disp_field and evaluated in vector_field."""
    dg = [tdgf(x) for tdgf in torch_template_deformed_grad_factory(len(x))]
    v = [vff(x) for vff in vector_field_factory(len(x))]
    return sum(dgi * vi for dgi, vi in zip(dg, v))


def numpy_inv_deformed_template(x):
    """Analytic inverse deformation of the template function."""
    disp = [dsp(x) for dsp in disp_field_factory(len(x))]
    disp_x = [xi - di for xi, di in zip(x, disp)]
    return numpy_template_function(disp_x)

def torch_inv_deformed_template(x):
    """Analytic inverse deformation of the template function."""
    disp = [dsp(x) for dsp in disp_field_factory(len(x))]
    disp_x = [xi - di for xi, di in zip(x, disp)]
    return torch_template_function(disp_x)


# --- LinDeformFixedTempl --- #


def test_fixed_templ_init(odl_impl_device_pairs):
    """Test init and props of linearized deformation with fixed template."""
    impl, device = odl_impl_device_pairs
    space = odl.uniform_discr(0, 1, 5, impl=impl,device=device)

    if impl == 'numpy':
        template_function = numpy_template_function 
    else:
        template_function = torch_template_function

    template = space.element(template_function)

    # Valid input
    op = LinDeformFixedTempl(template)
    assert repr(op) != ''
    op = LinDeformFixedTempl(template, domain=space.astype('float32') ** 1)
    assert repr(op) != ''

    # Invalid input
    with pytest.raises(TypeError):
        # template_function not a DiscretizedSpaceElement
        LinDeformFixedTempl(template_function)

@pytest.fixture
def space(odl_impl_device_pairs):
    """Provide a space for unit tests."""
    impl, device = odl_impl_device_pairs
    ndim = 2
    # supported_dtypes = odl.lookup_array_backend(impl).available_dtypes
    # # if np.dtype(dtype) not in supported_dtypes:
    # #     pytest.skip('dtype not available for this backend')

    return odl.uniform_discr(
        [-1] * ndim, [1] * ndim, [20] * ndim, impl=impl, device=device
    )


def test_fixed_templ_call(space, interp):
    """Test call of linearized deformation with fixed template."""
    # Define the analytic template as the hat function and its gradient
    if space.impl == 'numpy':
        template_function = numpy_template_function 
        deformed_template = numpy_deformed_template
    else:
        if space.dtype_identifier == 'complex128':
            return 
        template_function = torch_template_function
        deformed_template = torch_deformed_template

    template = space.element(template_function)
    deform_op = LinDeformFixedTempl(template, interp=interp)

    # Calculate result and exact result
    true_deformed_templ = space.element(deformed_template)

    deformed_templ = deform_op(disp_field_factory(space.ndim))

    # Verify that the result is within error limits
    error = (true_deformed_templ - deformed_templ).norm()
    rlt_err = error / deformed_templ.norm()
    assert rlt_err < error_bound(interp)


def test_fixed_templ_deriv(space, interp):
    """Test derivative of linearized deformation with fixed template."""
    if not space.is_real:
        pytest.skip('derivative not implemented for complex dtypes')

    if space.impl == 'numpy':
        template_function = numpy_template_function 
        fixed_templ_deriv = numpy_fixed_templ_deriv
    else:
        template_function = torch_template_function
        fixed_templ_deriv = torch_fixed_templ_deriv

    # Set up template and displacement field
    template = space.element(template_function)
    disp_field = disp_field_factory(space.ndim)
    vector_field = vector_field_factory(space.ndim)
    fixed_templ_op = LinDeformFixedTempl(template, interp=interp)

    # Calculate result
    fixed_templ_op_deriv = fixed_templ_op.derivative(disp_field)
    fixed_templ_deriv_comp = fixed_templ_op_deriv(vector_field)

    # Calculate the analytic result
    fixed_templ_deriv_exact = space.element(fixed_templ_deriv)

    # Verify that the result is within error limits
    error = (fixed_templ_deriv_exact - fixed_templ_deriv_comp).norm()
    rlt_err = error / fixed_templ_deriv_comp.norm()
    assert rlt_err < error_bound(interp)


# --- LinDeformFixedDisp --- #


def test_fixed_disp_init(odl_impl_device_pairs):
    """Test init and props of lin. deformation with fixed displacement."""
    impl, device=odl_impl_device_pairs
    space = odl.uniform_discr(0, 1, 5, impl=impl, device=device)
    disp_field = space.tangent_bundle.element(
        disp_field_factory(space.ndim))

    # Valid input
    op = LinDeformFixedDisp(disp_field)
    assert repr(op) != ''
    op = LinDeformFixedDisp(disp_field, templ_space=space)
    assert repr(op) != ''

    # Non-valid input
    with pytest.raises(TypeError):  # displacement not ProductSpaceElement
        LinDeformFixedDisp(space.one())
    with pytest.raises(TypeError):  # templ_space not DiscretizedSpace
        LinDeformFixedDisp(disp_field, space.tangent_bundle)
    with pytest.raises(TypeError):  # templ_space not a power space
        bad_pspace = odl.ProductSpace(space, odl.rn(3))
        LinDeformFixedDisp(disp_field, bad_pspace)
    with pytest.raises(TypeError):  # templ_space not based on DiscretizedSpace
        bad_pspace = odl.ProductSpace(odl.rn(2), 1)
        LinDeformFixedDisp(disp_field, bad_pspace)
    with pytest.raises(TypeError):  # wrong dtype on templ_space
        wrong_dtype = odl.ProductSpace(space.astype(complex), 1)
        LinDeformFixedDisp(disp_field, wrong_dtype)
    with pytest.raises(ValueError):  # vector field spaces don't match
        bad_space = odl.uniform_discr(0, 1, 10)
        LinDeformFixedDisp(disp_field, bad_space)


def test_fixed_disp_call(space, interp):
    """Test call of lin. deformation with fixed displacement."""
    if space.impl == 'numpy':
        template_function = numpy_template_function 
        deformed_template = numpy_deformed_template
    else:
        if space.dtype_identifier == 'complex128':
            return 
        template_function = torch_template_function
        deformed_template = torch_deformed_template

    template = space.element(template_function)
    disp_field = space.real_space.tangent_bundle.element(
        disp_field_factory(space.ndim))

    # Calculate result and exact result
    deform_op = LinDeformFixedDisp(
        disp_field, templ_space=space, interp=interp
    )
    deformed_templ = deform_op(template)
    true_deformed_templ = space.element(deformed_template)

    # Verify that the result is within error limits
    error = (true_deformed_templ - deformed_templ).norm()
    rlt_err = error / deformed_templ.norm()
    assert rlt_err < error_bound(interp)


def test_fixed_disp_inv(space, interp):
    """Test inverse of lin. deformation with fixed displacement."""
    if space.impl == 'numpy':
        template_function = numpy_template_function 
    else:
        if space.dtype_identifier == 'complex128':
            return 
        template_function = torch_template_function
        
    # Set up template and displacement field
    template = space.element(template_function)
    disp_field = space.real_space.tangent_bundle.element(
        disp_field_factory(space.ndim))

    # Verify that the inverse is in fact a (left and right) inverse
    deform_op = LinDeformFixedDisp(
        disp_field, templ_space=space, interp=interp
    )

    result_op_inv = deform_op(deform_op.inverse(template))
    error = (result_op_inv - template).norm()
    rel_err = error / template.norm()
    assert rel_err < 2 * error_bound(interp)  # need a bit more tolerance

    result_inv_op = deform_op.inverse(deform_op(template))
    error = (result_inv_op - template).norm()
    rel_err = error / template.norm()
    assert rel_err < 2 * error_bound(interp)  # need a bit more tolerance


def test_fixed_disp_adj(space, interp):
    """Test adjoint of lin. deformation with fixed displacement."""
    # Set up template and displacement field

    if space.impl == 'numpy':
        template_function = numpy_template_function 
        inv_deformed_template = numpy_inv_deformed_template
        exp_div_inv_disp = numpy_exp_div_inv_disp
    else:
        if space.dtype_identifier == 'complex128':
            return 
        template_function = torch_template_function
        inv_deformed_template = torch_inv_deformed_template
        exp_div_inv_disp = torch_exp_div_inv_disp

    template = space.element(template_function)
    disp_field = space.real_space.tangent_bundle.element(
        disp_field_factory(space.ndim))

    # Calculate result
    deform_op = LinDeformFixedDisp(
        disp_field, templ_space=space, interp=interp
    )
    deformed_templ_adj = deform_op.adjoint(template)

    # Calculate the analytic result
    true_deformed_templ_adj = space.element(inv_deformed_template)
    exp_div = space.element(exp_div_inv_disp)
    true_deformed_templ_adj *= exp_div

    # Verify that the result is within error limits
    error = (deformed_templ_adj - true_deformed_templ_adj).norm()
    rel_err = error / true_deformed_templ_adj.norm()
    assert rel_err < error_bound(interp)

    # Verify the adjoint definition <Ax, x> = <x, A^* x>
    deformed_templ = deform_op(template)
    inner1 = deformed_templ.inner(template)
    inner2 = template.inner(deformed_templ_adj)
    assert inner1 == pytest.approx(inner2, abs=.1)


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Douglas-Rachford splitting algorithm for convex optimization."""

from __future__ import print_function, division, absolute_import

import numpy as np

from odl.core.operator import Operator


__all__ = ('douglas_rachford_pd', 'douglas_rachford_pd_stepsize')


def douglas_rachford_pd(x, f, g, L, niter, tau=None, sigma=None,
                        callback=None, **kwargs):
    r"""Douglas-Rachford primal-dual splitting algorithm.

    Minimizes the sum of several convex functions composed with linear
    operators::

        min_x f(x) + sum_i g_i(L_i x)

    where ``f``, ``g_i`` are convex functions, ``L_i`` are linear `Operator`'s.

    Can also be used to solve the more general problem::

        min_x f(x) + sum_i (g_i @ l_i)(L_i x)

    where ``l_i`` are convex functions and ``@`` is the infimal convolution::

        (g @ l)(x) = inf_y g(y) + l(x - y)

    For references on the algorithm, see algorithm 3.1 in [BH2013].

    Parameters
    ----------
    x : `LinearSpaceElement`
        Initial point, updated in-place.
    f : `Functional`
        `proximal factory` for the function ``f``.
    g : sequence of `Functional`'s
        Sequence of of the functions ``g_i``. Needs to have
        ``g[i].convex_conj.proximal``.
    L : sequence of `Operator`'s
        Sequence of `Operator`'s with as many elements as ``g``.
    niter : int
        Number of iterations.
    tau : float, optional
        Step size parameter for ``f``.
        Default: Sufficient for convergence, see
        `douglas_rachford_pd_stepsize`.
    sigma : sequence of floats, optional
        Step size parameters for the ``g_i``'s.
        Default: Sufficient for convergence, see
        `douglas_rachford_pd_stepsize`.
    callback : callable, optional
        Function called with the current iterate after each iteration.

    Other Parameters
    ----------------
    l : sequence of `Functional`'s, optional
        Sequence of of the functions ``l_i``. Needs to have
        ``l[i].convex_conj.proximal``.
        If omitted, the simpler problem without ``l_i``  will be considered.
    lam : float or callable, optional
        Overrelaxation step size. If callable, it should take an index
        (starting at zero) and return the corresponding step size.

    Notes
    -----
    The mathematical problem to solve is

    .. math::
       \min_x f(x) + \sum_{i=0}^n (g_i \Box l_i)(L_i x),

    where :math:`f`, :math:`g_i`, :math:`l_i` are proper, convex and lower
    semicontinuous and :math:`L_i` are linear operators. The infimal
    convolution :math:`g \Box l` is defined by

    .. math::
       (g \Box l)(x) = \inf_y g(y) + l(x - y).

    The simplified problem,

    .. math::
        \min_x f(x) + \sum_{i=0}^n g_i(L_i x),

    can be obtained by setting

    .. math::
        l(x) = 0 \text{ if } x = 0, \infty \text{ else.}

    To guarantee convergence, the parameters :math:`\tau`, :math:`\sigma_i`
    and :math:`L_i` need to satisfy

    .. math::
       \tau \sum_{i=1}^n \sigma_i \|L_i\|^2 < 4

    The parameter :math:`\lambda` needs to satisfy :math:`0 < \lambda < 2`
    and if it is given as a function it needs to satisfy

    .. math::
        \sum_{n=1}^\infty \lambda_n (2 - \lambda_n) = +\infty.

    See Also
    --------
    odl.solvers.nonsmooth.primal_dual_hybrid_gradient.pdhg :
        Solver for similar problems.
    odl.solvers.nonsmooth.forward_backward.forward_backward_pd :
        Solver for similar problems which can additionaly handle a
        differentiable term.

    References
    ----------
    [BH2013] Bot, R I, and Hendrich, C. *A Douglas-Rachford type
    primal-dual method for solving inclusions with mixtures of
    composite and parallel-sum type monotone operators*. SIAM Journal
    on Optimization, 23.4 (2013), pp 2541--2565.
    """
    # Validate input
    m = len(L)
    if not all(isinstance(op, Operator) for op in L):
        raise ValueError('`L` not a sequence of operators')
    if not all(op.is_linear for op in L):
        raise ValueError('not all operators in `L` are linear')
    if not all(x in op.domain for op in L):
        raise ValueError('`x` not in the domain of all operators')
    if len(g) != m:
        raise ValueError('len(prox_cc_g) != len(L)')

    tau, sigma = douglas_rachford_pd_stepsize(L, tau, sigma)

    if len(sigma) != m:
        raise ValueError('len(sigma) != len(L)')

    prox_cc_g = [gi.convex_conj.proximal for gi in g]

    # Get parameters from kwargs
    l = kwargs.pop('l', None)
    if l is not None and len(l) != m:
        raise ValueError('`l` does not have the same number of '
                         'elements as `L`')
    if l is not None:
        prox_cc_l = [li.convex_conj.proximal for li in l]

    lam_in = kwargs.pop('lam', 1.0)
    if not callable(lam_in) and not (0 < lam_in < 2):
        raise ValueError('`lam` must callable or a number between 0 and 2')
    lam = lam_in if callable(lam_in) else lambda _: lam_in

    # Check for unused parameters
    if kwargs:
        raise TypeError('got unexpected keyword arguments: {}'.format(kwargs))

    # Pre-allocate values
    v = [Li.range.zero() for Li in L]
    p1 = x.space.zero()
    p2 = [Li.range.zero() for Li in L]
    z1 = x.space.zero()
    # Save a bit of memory: z2 elements are local to the loop where they
    # are used
    rans = {Li.range for Li in L}
    z2 = {ran: ran.zero() for ran in rans}
    w1 = x.space.zero()
    w2 = [Li.range.zero() for Li in L]

    for k in range(niter):
        lam_k = lam(k)

        if len(L) > 0:
            # Compute z1 = sum(Li.adjoint(vi) for Li, vi in zip(L, v))
            # NB: we abuse z1 as temporary here, in contrast to the algorithm
            # in the paper
            L[0].adjoint(v[0], out=z1)
            for Li, vi in zip(L[1:], v[1:]):
                Li.adjoint(vi, out=p1)
                z1 += p1

            z1.lincomb(1, x, -tau / 2, z1)
        else:
            z1.assign(x)

        f.proximal(tau)(z1, out=p1)
        # Now p1 = prox[tau*f](x - tau/2 * sum(Li^* vi))
        # Temporary z1 is no longer needed

        # w1 = 2 * p1 - x
        w1.lincomb(2, p1, -1, x)

        # Part 1 of x += lam(k) * (z1 - p1)
        x.lincomb(1, x, -lam_k, p1)

        # Now p1 is free to use as temporary; however, since p1 holds the
        # current primal iterate (not x) we call the callback here already
        # and return early if we're in the last iteration (also saves some
        # computation)
        if callback is not None:
            callback(p1)
        if k == niter - 1:
            x.assign(p1)
            return

        for i in range(m):
            # Compute p2[i] = prox[sigma * g^*](v[i] + sigma[i]/2 * L[i](w1))
            L[i](w1, out=p2[i])
            p2[i].lincomb(1, v[i], sigma[i] / 2, p2[i])
            prox_cc_g[i](sigma[i])(p2[i], out=p2[i])
            # w2[i] = 2 * p2[i] - v[i]
            w2[i].lincomb(2, p2[i], -1, v[i])

        if len(L) > 0:
            # Compute p1 = sum(Li.adjoint(w2i) for Li, w2i in zip(L, w2))
            # NB: we abuse p1 as temporary here, in contrast to the algorithm
            # in the paper
            L[0].adjoint(w2[0], out=p1)
            for Li, w2i in zip(L[1:], w2[1:]):
                Li.adjoint(w2i, out=z1)
                p1 += z1
        else:
            p1.set_zero()

        # z1 = w2 - tau/2 * p1
        z1.lincomb(1, w1, -tau / 2, p1)

        # Part 2 of x += lam(k) * (z1 - p1)
        x.lincomb(1, x, lam_k, z1)

        # p1 = 2 * z1 - w1
        p1.lincomb(2, z1, -1, w1)
        for i in range(m):
            z2i = z2[L[i].range]
            # Compute
            # z2[i] = prox[sigma[i] * l[i]^*](w2[i] + sigma[i]/2 * L[i](p1))
            L[i](p1, out=z2i)
            z2i.lincomb(1, w2[i], sigma[i] / 2, L[i](p1))
            # prox_cc_l is the identity if `l is None`, thus omitted in that
            # case
            if l is not None:
                prox_cc_l[i](sigma[i])(z2i, out=z2i)

            # Compute v[i] += lam(k) * (z2[i] - p2[i])
            v[i].lincomb(1, v[i], lam_k, z2i)
            v[i].lincomb(1, v[i], -lam_k, p2[i])


def _operator_norms(L):
    """Get operator norms if needed.

    Parameters
    ----------
    L : sequence of `Operator` or float
        The operators or the norms of the operators that are used in the
        `douglas_rachford_pd` method. For `Operator` entries, the norm
        is computed with ``Operator.norm(estimate=True)``.
    """
    L_norms = []
    for Li in L:
        if np.isscalar(Li):
            L_norms.append(float(Li))
        elif isinstance(Li, Operator):
            L_norms.append(Li.norm(estimate=True))
        else:
            raise TypeError('invalid entry {!r} in `L`'.format(Li))
    return L_norms


def douglas_rachford_pd_stepsize(L, tau=None, sigma=None):
    r"""Default step sizes for `douglas_rachford_pd`.

    Parameters
    ----------
    L : sequence of `Operator` or float
        The operators or the norms of the operators that are used in the
        `douglas_rachford_pd` method. For `Operator` entries, the norm
        is computed with ``Operator.norm(estimate=True)``.
    tau : positive float, optional
        Use this value for ``tau`` instead of computing it from the
        operator norms, see Notes.
    sigma : tuple of float, optional
        The ``sigma`` step size parameters for the dual update.

    Returns
    -------
    tau : float
        The ``tau`` step size parameter for the primal update.
    sigma : tuple of float
        The ``sigma`` step size parameters for the dual update.

    Notes
    -----
    To guarantee convergence, the parameters :math:`\tau`, :math:`\sigma_i`
    and :math:`L_i` need to satisfy

    .. math::
       \tau \sum_{i=1}^n \sigma_i \|L_i\|^2 < 4.

    This function has 4 options, :math:`\tau`/:math:`\sigma` given or not
    given.

    - If neither :math:`\tau` nor :math:`\sigma` are given, they are chosen as:

        .. math::
            \tau = \frac{1}{\sum_{i=1}^n \|L_i\|},
            \quad
            \sigma_i = \frac{2}{n \tau \|L_i\|^2}

    - If only :math:`\sigma` is given, :math:`\tau` is set to:

        .. math::
            \tau = \frac{2}{\sum_{i=1}^n \sigma_i \|L_i\|^2}

    - If only :math:`\tau` is given, :math:`\sigma` is set
      to:

        .. math::
            \sigma_i = \frac{2}{n \tau \|L_i\|^2}

    - If both are given, they are returned as-is without further validation.
    """
    if tau is None and sigma is None:
        L_norms = _operator_norms(L)

        tau = 1 / sum(L_norms)
        sigma = [2.0 / (len(L_norms) * tau * Li_norm ** 2)
                 for Li_norm in L_norms]

        return tau, tuple(sigma)
    elif tau is None:
        L_norms = _operator_norms(L)

        tau = 2 / sum(si * Li_norm ** 2
                      for si, Li_norm in zip(sigma, L_norms))
        return tau, tuple(sigma)
    elif sigma is None:
        L_norms = _operator_norms(L)

        tau = float(tau)
        sigma = [2.0 / (len(L_norms) * tau * Li_norm ** 2)
                 for Li_norm in L_norms]

        return tau, tuple(sigma)
    else:
        return float(tau), tuple(sigma)
